import mock
import pytest

from gantry.query import main
from gantry.query.client import GantryQuery


@pytest.mark.parametrize("backend", ["", "/some/file", "tcp://other/protocol"])
def test_global_init_error(backend):
    with mock.patch.dict("os.environ", {"GANTRY_LOGS_LOCATION": backend}):
        with pytest.raises(ValueError):
            main.init()


@pytest.mark.parametrize("valid_backend", ["https://app.gantry.io", "http://app.gantry.io"])
def test_global_init_valid_logs_location(valid_backend):
    with mock.patch.dict("os.environ", {"GANTRY_LOGS_LOCATION": valid_backend}):
        main.init(api_key="ABCD1234")


def test_no_api_key_provided():
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": ""}):
        with pytest.raises(ValueError):
            main.init(api_key=None)


@mock.patch("gantry.query.main.globals")
def test_passed_api_key_overrides_env(mock_globals):
    passed_api_key = "some_key"
    env_api_key = "some_other_key"
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": env_api_key}):
        main.init(passed_api_key)
        assert mock_globals._Query._api_client._api_key == passed_api_key
        assert mock_globals._Query._api_client._api_key != env_api_key


@mock.patch("gantry.query.main.globals")
def test_global_init_default(mock_globals):
    passed_api_key = "some_key"
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": passed_api_key}):
        main.init()  # api_key is required as param or env var
        assert mock_globals._Query._api_client._host == "https://app.gantry.io"
        assert mock_globals._Query._api_client._api_key == passed_api_key


@mock.patch("gantry.query.main.globals")
def test_global_init(mock_globals):
    main.init(api_key="ABCD1234")
    assert mock_globals._Query._api_client._host == "https://app.gantry.io"
    assert mock_globals._Query._api_client._api_key == "ABCD1234"


@pytest.mark.parametrize(
    ["method", "params"],
    [
        ("list_application_environments", {"application": "foobar"}),
        ("list_application_versions", {"application": "foobar"}),
        (
            "list_application_views",
            {"application": "foobar", "version": "1.2.3", "environment": "dev"},
        ),
        (
            "list_application_batches",
            {"application": "foobar", "version": "1.2.3"},
        ),
        ("list_applications", {}),
        ("print_application_info", {"application": "foobar"}),
        (
            "query",
            {
                "application": "foobar",
                "start_time": "2020",
                "end_time": "2020",
                "version": "1.2.3",
                "environment": "dev",
                "filters": [],
                "view": "foo",
                "tags": {"foo": "bar"},
                "batch_id": "barbaz",
            },
        ),
        (
            "create_view",
            {
                "name": "barbaz",
                "application": "foobar",
                "version": "1.2.3",
                "start_time": "2020",
                "end_time": "2020",
                "duration": "1D",
                "data_filters": [],
                "tag_filters": {"foo": "bar"},
            },
        ),
    ],
)
@mock.patch("gantry.query.main.validate_init")
@mock.patch("gantry.query.main.globals._Query", spec=GantryQuery)
def test_query_methods(mock_query, mock_validate_init, method, params):
    getattr(mock_query, method).return_value = "foo"
    assert getattr(main, method)(**params) == "foo"
    mock_validate_init.assert_called_once()
    getattr(mock_query, method).assert_called_once_with(**params)



import datetime
import logging
import random
from collections import defaultdict
from typing import Dict, List, Optional, Union

import isodate
from halo import Halo
from tabulate import tabulate

from gantry.api_client import APIClient
from gantry.exceptions import GantryRequestException
from gantry.query.core.dataframe import GantryDataFrame
from gantry.query.core.distance import GantryDistance
from gantry.query.core.metric import GantryMetric
from gantry.query.core.utils import (
    _build_batch_id_filter,
    _get_application_batches,
    get_application_node_id,
    get_application_views,
)

logger = logging.getLogger(__name__)

COLOR_OPTIONS_FOR_VIEW = [
    "#32D3EC",
    "#3998D9",
    "#126196",
    "#B184FE",
    "#7C3AED",
    "#4E16AD",
    "#FDC17D",
    "#FF8900",
    "#FFE47A",
    "#FACC15",
    "#6EE7B7",
    "#CDDC39",
    "#389D3C",
    "#8BC34A",
    "#D2AFC5",
    "#904C77",
    "#642F50",
    "#607D8B",
]


class GantryQuery:
    def __init__(self, api_client: APIClient) -> None:
        self._api_client = api_client
        self.distance = GantryDistance(self._api_client)
        self.metric = GantryMetric(self._api_client)

    def list_applications(self) -> List[str]:
        """
        List all applications in Gantry.

        Returns:
            List of application names that are stored in gantry.
        """
        query_params = {"limit": 10}
        response = self._api_client.request(
            "GET", "/api/v1/models", params=query_params, raise_for_status=True
        )
        return list(response["data"].keys())

    def list_application_versions(self, application: str) -> List[str]:
        """
        List the stored versions integer and version names for a given application.

        The version integer is used in conjunction with the application name to uniquely identify
        the application and its events in Gantry.

        Args:
            application (str): application name string.
        Returns:
            List of versions for |application| that are stored in gantry.
        """
        version_response = self._api_client.request(
            "GET",
            "/api/v1/models/" + application,
            params={"include_names": True},
            raise_for_status=True,
        )
        data = version_response["data"]
        return [version_dict["version"] for version_dict in data["versions"]]

    def list_application_environments(self, application: str) -> List[str]:
        """
        List the available environments for a given application.

        Args:
            application (str): application name string.
        Returns:
            List of environments
        """
        response = self._api_client.request(
            "GET", "/api/v1/models/" + application, raise_for_status=True
        )
        return response["data"]["environments"]

    def query(
        self,
        application: str,
        start_time: Optional[Union[str, datetime.datetime]] = None,
        end_time: Optional[Union[str, datetime.datetime]] = None,
        version: Optional[Union[int, str]] = None,
        environment: Optional[str] = None,
        filters: Optional[List[Dict]] = None,
        view: Optional[str] = None,
        tags: Optional[dict] = None,
        batch_id: Optional[str] = None,
    ) -> GantryDataFrame:
        """
        Query for a window of data for an application and version with a given start, end
        time.

        Args:
            application (str): application name string.
            start_time (str or datetime): datetime object for start time of the window.
                If not provided, defaults to previous 24 hours
            end_time (str or datetime): datetime object for the end time of the window.
                If not provided, defaults to now
            version (optional, int or str): application version.
            environment: (optional, str) - environment string to pull application for.
                This parameter takes precedence over a tags["env"] value in case
                tags are passed.
            filters: (optional, list[dict]) - optional filters for the query. Use this
                parameter only if you are familiar with the dict representation of
                the filters.
            view: (optional, str) - The view name to query data from. A view is a saved
                configuration of filters/start_time/end_time, so this parameter cannot
                be used with start_time, end_time, filters or batch_id parameters.
            tags: (optional, dict) - Optional tags to filter data. Each record has tags
                associated that were provided at ingestion time. The environment parameter
                overrides an "env" tag.
            batch_id: (optional, str) - The batch_id to query data from. Only data ingested
                in the given batch will be retrieved. This parameter cannot be used with
                'view' parameter.

        Returns:
            Gantry Dataframe with the data.
        """
        if view is not None:
            if not (
                start_time is None and end_time is None and filters is None and batch_id is None
            ):
                raise ValueError(
                    "Cannot provide 'view' parameter with "
                    "start_time/end_time/filters/batch_id at the same time"
                )

            views = self.list_application_views(
                application, version=version, environment=environment
            )
            if view not in views:
                raise ValueError(
                    f"View '{view}' not found. Available views for this version/env: {views}"
                )

            return GantryDataFrame.from_view(
                self._api_client, application, view, version, environment, tags=tags
            )

        if batch_id:
            batches = _get_application_batches(self._api_client, application, version=version)
            batch_ids = [b["id"] for b in batches]
            if batch_id not in batch_ids:
                raise ValueError(
                    f"Batch '{batch_id}' not found. Available batches for this version: {batch_ids}"
                )

            filters = filters or []
            filters.append(_build_batch_id_filter(batch_id))

        start_time = start_time or "-24H"
        end_time = end_time or "now"

        return GantryDataFrame(
            self._api_client,
            application,
            version=version,
            env=environment,
            start_time=start_time,
            end_time=end_time,
            filters=filters,
            tags=tags,
        )

    def get_current_feedback_schema(self, application: str):
        """Returns current feedback schema for the provided application."""
        raise NotImplementedError()

    def update_feedback_schema(
        self, application, feedback_fields: List[Dict[str, str]], create: bool = True
    ):
        """Update feedback schema with feedback_fields for the provided application

        This function updates the feedback schema to *only* have the feedback fields specified in
        the `feedback_fields` argument. If you want to add new feedback fields to an existing
        feedback schema, use the `add_feedback_field` function.

        Args:
            application: The name of the relevant application to apply this update for.
            feedback_fields: A list of dicts with keys 'name' and 'type'. For example:

                .. code-block:: python

                    [{'name': 'label', 'type': 'int'}, {'name': 'label2', 'type': 'str'}]

            create: If create=True, then if a feedback schema doesn't exist for the application
                then this function will create a new schema. If create=False, then if a feedback
                schema doesn't already exist, will raise an error.
        """
        raise NotImplementedError()

    def add_feedback_field(self, application: str, feedback_field: Dict[str, str]) -> int:
        """Adds a feedback field to current feedback schema.

        A current feedback schema must exist to use this method, otherwise it will raise an error.
        If it doesn't, use the `update_feedback_schema` method to generate a new one.

        Args:
            application: The name of the application apply this update to.
            feedback_field: A dict with keys 'name' and 'type' specifying the `name` and `type` of
                the feedback field. For example:

                .. code-block:: python

                    {'name': 'label', 'type': 'int'}

        """
        raise NotImplementedError()

    def get_current_metric_schema(self, application: str):
        """Returns current metrics schema for application with name=application."""
        raise NotImplementedError()

    def update_metric_schema(self, application: str, metrics: List[dict], create: bool = True):
        """Update metrics schema with metrics for application.

        This function updates the metrics schema to *only* have the metrics specified in
        the `metrics` argument. If you want to add new metrics to an existing
        feedback schema, use the `add_metric` function. Must already have schemas for
        prediction events and feedback events in place before created a metric schema.

        Args:
            application: The name of the relevant application to apply this update for.
            metrics: A list of dicts, where each dict specifies a metric. For example a metric
                dict looks like:

                .. code-block:: python

                    {
                        'name': 'my_metric',
                        'output_features': 'outputs',
                        'feedback_fields': 'feedback.label',
                        'metric_fn': 'mse'
                    }

            create: If create=True, then if a metric schema doesn't exist for the func_name
                then this function will create a new schema. If create=False, then if a metric
                schema doesn't already exist, will raise an error.
        """
        raise NotImplementedError()

    def add_metric(self, application: str, metric: dict) -> int:
        """Adds a metric to current metric schema.

        A current metric schema must exist to use this method, otherwise it will raise an error.
        If it doesn't, use the `update_metric_schema` method to generate a new one.

        Args:
            application (str): The name of the relevant application to apply this update for.
            metric: A dict that specifs a metric. For example:

                .. code-block:: python

                    {
                        'name': 'my_metric',
                        'output_features': 'outputs',
                        'feedback_fields': 'feedback.label',
                        'metric_fn': 'mse'
                    }

        """
        raise NotImplementedError()

    def create_view(
        self,
        name: str,
        application: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        duration: Optional[datetime.timedelta] = None,
        version: Optional[str] = None,
        tag_filters: Optional[Dict[str, str]] = None,
        data_filters: Optional[List[Dict]] = None,
    ) -> None:
        """Create a view for an application.

        Args:
            name (str): The name of the view.
            application (str): The application in which the view will be created.
            start_time (optional, datetime): Start time of the view. If you pass start_time,
                you need to specify end_time and cannot set duration.
            end_time (optional, datetime): End time of the view. If you pass end_time, you need to
                specify start_time also and cannot set duration.
            duration (optional, timedelta): Time window of the view. End time will be present time.
                For example, timedelta(days=1) will look 24hs back.
                IMPORTANT: if exact datetimes want to be provided, use start_time/end_time instead.
            version (optional, str): The version of the app that the view should use.
            tag_filters (optional, dict): A dictionary with tags to filter the data of the view.
            data_filters (optional, list[dict]): A list of Gantry-formatted filters. For example:

                .. code-block:: python

                    [
                        {
                            'feature_name': 'inputs.loan_amount',
                            'lower_bound': '100',
                            'upper_bound': '200',
                        }
                    ]

        """
        data = {
            "name": name,
            "tag_filters": tag_filters or {},
            "data_filters": data_filters or [],
            "details": {"color": random.choice(COLOR_OPTIONS_FOR_VIEW)},
        }

        if duration:
            if start_time or end_time:
                raise ValueError("Cannot provide duration and start_time/end_time at the same time")

            data["duration"] = isodate.duration_isoformat(duration)

        if start_time or end_time:
            if not (start_time and end_time):
                raise ValueError("Need to provide both start_time and end_time")

            data["start_time"] = start_time  # type: ignore
            data["end_time"] = end_time  # type: ignore

        if not (duration or start_time or end_time):
            raise ValueError("Need to provide duration or start_time/end_time")

        try:
            model_node_id = get_application_node_id(self._api_client, application, version=version)
            data["model_node_id"] = model_node_id
        except GantryRequestException:
            raise ValueError(
                f"Application {application}[{version if version else 'latest'}] does not exist."
            )

        self._api_client.request("POST", "/api/v1/views", json=data, raise_for_status=True)
        return None

    def list_application_views(
        self,
        application: str,
        version: Optional[Union[str, int]] = None,
        environment: Optional[str] = None,
    ) -> List[str]:
        """
        List saved views for given application, version and enviroment.

        Args:
            application (str): The name of the application.
            version (str): The version of the application to show the views
                from. If not provided, it will use the latest version.
            environment (str): The environment of the application to filter
                the views. If not provided, views from all environments will
                be shown.

        Returns:
            List of saved views names.
        """

        views = get_application_views(self._api_client, application, version, environment)
        return [view["name"] for view in views]

    def list_application_batches(
        self,
        application: str,
        version: Optional[Union[str, int]] = None,
    ) -> List[dict]:
        """
        List data batches for given application and version.

        Args:
            application (str): The name of the application.
            version (str): The version of the application to show the views
                from. If not provided, it will use the latest version.

        Returns:
            List of batches
        """
        return _get_application_batches(self._api_client, application, version)

    def print_application_info(self, application: str) -> None:
        """
        Prints a high-level report of an application.

        Args:
            application (str): The name of the application.

        """
        versions = self.list_application_versions(application)
        envs = self.list_application_environments(application)

        print("Available versions:")
        print("-------------------")
        print(versions)
        print()
        print("Available environments:")
        print("-----------------------")
        print(sorted(envs))
        print()
        print("Available views:")
        print("----------------")

        with Halo(text="Fetching views...", spinner="dots"):
            content = defaultdict(list)
            for version in versions:
                if self.list_application_views(application, version):
                    for env in envs:
                        views = get_application_views(self._api_client, application, version, env)
                        for view in views:
                            content["name"].append(view["name"])
                            if view.get("duration"):
                                content["duration"].append(isodate.parse_duration(view["duration"]))
                                content["from"].append("null")
                                content["to"].append("null")
                            else:
                                content["duration"].append("null")
                                content["from"].append(
                                    datetime.datetime.fromisoformat(view["start_time"])
                                )
                                content["to"].append(
                                    datetime.datetime.fromisoformat(view["end_time"])
                                )
                            content["version"].append(version)
                            content["env"].append(env)
        print(tabulate(content, headers="keys"))



from gantry.alerts.client import AlertsAggregation, AlertsCheck
from gantry.alerts.main import (  # noqa: F401
    create_alert,
    create_slack_notification,
    delete_alert,
    delete_slack_notification,
    get_alerts,
    get_slack_notifications,
    update_alert,
)

__all__ = [
    "get_alerts",
    "create_alert",
    "update_alert",
    "create_slack_notification",
    "get_slack_notifications",
    "delete_alert",
    "delete_slack_notification",
    "AlertsAggregation",
    "AlertsCheck",
]



import pytest

from gantry.exceptions import GantryLoggingDataTypeError
from gantry.validators import validate_logged_field_datatype


def test_validator_list_not_primitive():
    # does not accept arbitrary malformed
    with pytest.raises(GantryLoggingDataTypeError):
        validate_logged_field_datatype(
            {"some-invalid-input": ["invalid", {"invalid": True}], "some-input": True}
        )


def test_validator_list_3d():
    # Does not accept 3 dim array
    with pytest.raises(GantryLoggingDataTypeError):
        validate_logged_field_datatype([[[1]], [[2]], [[3]]])


def test_validator_list_2d():
    # Does not raise for 2 dim array
    validate_logged_field_datatype([[1], [2], [3]])


def test_validator_list_1d():
    # Does not raise for 1 dim array
    validate_logged_field_datatype([1])


def test_validator_nested_dict_3():
    # Does not raise for 3 nested dicts
    validate_logged_field_datatype({"a": {"b": {"c": "d"}}})


def test_validator_nested_dict_2():
    # Does not raise for 2 nested dict
    validate_logged_field_datatype({"a": {"b": "c"}})


def test_validator_nested_list():
    # Does not raise for valid list inside dict
    validate_logged_field_datatype({"a": {"b": ["c"]}})


def test_validator_primitives():
    # Does not raise for primitives
    try:
        validate_logged_field_datatype("str")
        validate_logged_field_datatype(1)
        validate_logged_field_datatype(1.1)
        validate_logged_field_datatype(True)
        validate_logged_field_datatype(None)
    except GantryLoggingDataTypeError:
        pytest.fail("Raised GantryLoggingDataTypeError for a primitive")



import uuid

import boto3
import click.exceptions
import mock
import pytest
import responses
from botocore.exceptions import ClientError
from click.testing import CliRunner
from moto import mock_s3
from responses import matchers

from gantry.cli.labeling import (
    _validate_source,
    create,
    dashboard_export,
    dataset_export,
    list,
)
from gantry.exceptions import GantryRequestException

BUCKET_NAME = "test-bucket"
REGION = "us-west-2"
PROJECT_NAME = "test-project"
UNITTEST_HOST = "http://unittest"

LS_HOST = "http://test-ls-host"
LS_TOKEN = "test-ls-token"
LS_PROJ_ID = "fake-project-id"

AWS_ACCESS_KEY_ID = "test_access_key"
AWS_SECRET_ACCESS_KEY = "test_secret_access_key"
AWS_SESSION_TOKEN = "test_session_token"

TEST_ENV = {
    "GANTRY_API_KEY": "test",
    "GANTRY_LOGS_LOCATION": UNITTEST_HOST,
    "AWS_ACCESS_KEY_ID": AWS_ACCESS_KEY_ID,
    "AWS_SECRET_ACCESS_KEY": AWS_SECRET_ACCESS_KEY,
    "AWS_SESSION_TOKEN": AWS_SESSION_TOKEN,
    "AWS_DEFAULT_REGION": REGION,
}

unpatched_uuid = uuid.uuid4


def s3_setup(bucket_name=BUCKET_NAME, region=REGION):
    s3_client = boto3.client(
        "s3",
        region_name=region,
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        aws_session_token=AWS_SESSION_TOKEN,
    )
    try:
        s3_client.head_bucket(Bucket=bucket_name)
    except ClientError:
        s3_client.create_bucket(
            Bucket=bucket_name,
            CreateBucketConfiguration={"LocationConstraint": region},
        )

    return s3_client


def setup_resp_with_backoff(resp: responses.RequestsMock, *args, **kwargs):
    id = unpatched_uuid()
    final_json = kwargs.pop("json", {})
    final_json["response"] = "ok"
    final_json["status"] = "SUCCESS"

    resp.add(  # type: ignore
        *args,
        **kwargs,
        status=202,
        headers={"Location": f"test-location-{id}"},
    )
    resp.add(
        resp.GET,
        f"{UNITTEST_HOST}/test-location-{id}",
        json=final_json,
    )


def setup_resp_for_gantry_pings(resp: responses.RequestsMock):
    resp.add(resp.GET, f"{UNITTEST_HOST}/api/ping", status=200)
    resp.add(resp.GET, f"{UNITTEST_HOST}/api/v1/auth-ping", status=200)


def setup_resp_for_labelstudio(resp: responses.RequestsMock, s3_prefix: str):
    ls_params = {
        "bucket": BUCKET_NAME,
        "aws_access_key_id": AWS_ACCESS_KEY_ID,
        "aws_secret_access_key": AWS_SECRET_ACCESS_KEY,
        "aws_session_token": AWS_SESSION_TOKEN,
        "region_name": None,
        "s3_endpoint": None,
        "title": "Gantry S3 connection",
        "description": "Integration made by Gantry",
        "project": LS_PROJ_ID,
    }

    # create the project in label studio
    resp.add(
        resp.POST,
        f"{LS_HOST}/api/projects",
        status=201,
        json={"id": LS_PROJ_ID},
    )
    # register import s3
    resp.add(
        resp.POST,
        f"{LS_HOST}/api/storages/s3",
        match=[
            matchers.json_params_matcher(
                {
                    **ls_params,
                    "regex_filter": None,
                    "use_blob_urls": False,
                    "prefix": f"{s3_prefix}/source_data",
                    "presign": True,
                    "presign_ttl": 1,
                },
            )
        ],
        json={"id": "fake-import-id"},
    )
    # register export s3
    resp.add(
        resp.POST,
        f"{LS_HOST}/api/storages/export/s3",
        match=[
            matchers.json_params_matcher(
                {
                    **ls_params,
                    "can_delete_objects": False,
                    "prefix": f"{s3_prefix}/labeled_data",
                },
            )
        ],
        json={"id": "fake-export-id"},
    )
    # labelstudio import sync
    resp.add(
        resp.POST,
        f"{LS_HOST}/api/storages/s3/fake-import-id/sync",
        json={"response": "ok"},
    )


@mock.patch("uuid.uuid4", return_value="01234")
@mock.patch("gantry.cli.labeling.LabelStudioManager.check_status")
@mock.patch("gantry.cli.labeling._validate_dataset")
@mock_s3
def test_create_project_from_dataset_version(
    mock_validate_dataset, mock_ls_status, mock_uuid, datadir
):
    s3_client = s3_setup()
    runner = CliRunner()

    source = "gantry://datasets/fake-dataset/versions/fake-version"
    s3_prefix = f"{PROJECT_NAME}__gantry_{mock_uuid.return_value}"

    cli_args = [
        "--source",
        source,
        "--name",
        PROJECT_NAME,
        "--bucket",
        BUCKET_NAME,
        "--ls-token",
        LS_TOKEN,
        "--ls-host",
        LS_HOST,
        "--project-interface",
        str(datadir / "project_interface.xml"),
    ]

    with responses.RequestsMock() as resp:
        setup_resp_for_gantry_pings(resp)
        setup_resp_for_labelstudio(resp, s3_prefix)
        # make sure the bucket exists
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/customer-resources/buckets/{BUCKET_NAME}?storage_type=S3",
            json={},
        )
        # make sure the project doesn't exist
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/labeling/projects/{PROJECT_NAME}",
            status=404,
        )
        # create the project
        resp.add(
            resp.POST,
            f"{UNITTEST_HOST}/api/v1/labeling/projects",
            json={
                "data": {
                    "name": PROJECT_NAME,
                    "id": LS_PROJ_ID,
                },
                "response": "ok",
            },
            match=[
                matchers.json_params_matcher(
                    {
                        "name": PROJECT_NAME,
                        "bucket": BUCKET_NAME,
                        "s3_prefix": s3_prefix,
                        "source": source,
                        "ls_host": LS_HOST,
                        "import_storage_id": "fake-import-id",
                        "export_storage_id": "fake-export-id",
                    }
                )
            ],
        )
        # start the project
        setup_resp_with_backoff(
            resp,
            resp.PATCH,
            f"{UNITTEST_HOST}/api/v1/labeling/projects/{PROJECT_NAME}",
            match=[matchers.json_params_matcher({"new-status": "imported"})],
        )

        result = runner.invoke(create, args=cli_args, env=TEST_ENV)

        assert result.exit_code == 0

        assert "New Label Studio project created" in result.output
        assert "Registered labeling project in Gantry" in result.output
        assert f"Ready to start labeling: {LS_HOST}/projects" in result.output

        # make sure the folders were created
        objects = s3_client.list_objects_v2(Bucket=BUCKET_NAME, Prefix=s3_prefix)["Contents"]
        keys = [f["Key"] for f in objects]

        assert f"{s3_prefix}/source_data/" in keys
        assert f"{s3_prefix}/labeled_data/" in keys


@mock.patch("gantry.cli.labeling._validate_dataset")
@mock_s3
def test_create_project_duplicate_name(mock_validate_dataset, datadir):
    runner = CliRunner()
    source = "gantry://datasets/fake-dataset/versions/fake-version"

    cli_args = [
        "--source",
        source,
        "--name",
        PROJECT_NAME,
        "--bucket",
        BUCKET_NAME,
        "--ls-token",
        LS_TOKEN,
        "--ls-host",
        LS_HOST,
        "--project-interface",
        str(datadir / "project_interface.xml"),
    ]

    with responses.RequestsMock() as resp:
        setup_resp_for_gantry_pings(resp)
        # make sure the bucket exists
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/customer-resources/buckets/{BUCKET_NAME}?storage_type=S3",
            json={},
        )
        # make sure the project doesn't exist (but this time it does)
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/labeling/projects/{PROJECT_NAME}",
            status=200,
        )

        result = runner.invoke(create, args=cli_args, env=TEST_ENV)

        assert result.exit_code == 1

        assert "New Label Studio project created" not in result.output
        assert result.exc_info[0] == ValueError
        assert "Name must be unique" in result.exc_info[1].args[0]


@mock.patch("gantry.cli.labeling._validate_dataset")
@mock_s3
def test_create_project_no_bucket(mock_validate_dataset, datadir):
    runner = CliRunner()
    source = "gantry://datasets/fake-dataset/versions/fake-version"

    cli_args = [
        "--source",
        source,
        "--name",
        PROJECT_NAME,
        "--bucket",
        BUCKET_NAME,
        "--ls-token",
        LS_TOKEN,
        "--ls-host",
        LS_HOST,
        "--project-interface",
        str(datadir / "project_interface.xml"),
    ]

    with responses.RequestsMock() as resp:
        setup_resp_for_gantry_pings(resp)
        # make sure the bucket exists (but it doesn't)
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/customer-resources/buckets/{BUCKET_NAME}?storage_type=S3",
            status=404,
        )

        result = runner.invoke(create, args=cli_args, env=TEST_ENV)

        assert result.exit_code == 1

        assert "New Label Studio project created" not in result.output
        assert result.exc_info[0] == RuntimeError
        assert (
            "Labeling requires customers to register bucket with their secrets"
            in result.exc_info[1].args[0]
        )


@pytest.mark.parametrize(
    "created_new_dataset",
    [True, False],
)
def test_export_project_to_dataset(created_new_dataset):
    runner = CliRunner()

    dataset_name = "fake-dataset"
    dataset_id = "fake-dataset-id"
    dataset_json = {
        "name": dataset_name,
        "id": dataset_id,
        "bucket_name": BUCKET_NAME,
        "aws_region": REGION,
        "organization_id": "fake-org-id",
        "s3_prefix": "fake-prefix",
        "disabled": False,
    }

    cli_args = [
        "--name",
        PROJECT_NAME,
        "--destination",
        dataset_name,
        "--ls-token",
        LS_TOKEN,
    ]

    with responses.RequestsMock() as resp:
        setup_resp_for_gantry_pings(resp)
        # get project
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/labeling/projects/{PROJECT_NAME}",
            json={
                "data": {"export_storage_id": "fake-export-id", "ls_host": LS_HOST},
                "response": "ok",
            },
        )
        # labelstudio export sync
        resp.add(
            resp.POST,
            f"{LS_HOST}/api/storages/export/s3/fake-export-id/sync",
            json={"response": "ok"},
        )
        # setup dataset stuff
        if created_new_dataset:
            resp.add(
                resp.GET,
                f"{UNITTEST_HOST}/api/v1/datasets/{dataset_name}",
                status=404,
            )
            resp.add(
                resp.POST,
                f"{UNITTEST_HOST}/api/v1/datasets",
                json={"data": dataset_json, "response": "ok"},
                match=[matchers.json_params_matcher({"name": dataset_name})],
            )
        else:
            resp.add(
                resp.GET,
                f"{UNITTEST_HOST}/api/v1/datasets/{dataset_name}",
                json={"data": dataset_json, "response": "ok"},
            )
        # export labels to dataset
        setup_resp_with_backoff(
            resp,
            resp.POST,
            f"{UNITTEST_HOST}/api/v1/datasets/{dataset_id}/labeled-data",
            json={"id": "fake-version-id", "response": "ok"},
            match=[matchers.json_params_matcher({"labeling-project-name": PROJECT_NAME})],
        )

        result = runner.invoke(dataset_export, args=cli_args, env=TEST_ENV)

        assert result.exit_code == 0

        assert "Synced labeled data to S3 bucket" in result.output
        if created_new_dataset:
            assert "Created new dataset" in result.output
        else:
            assert "Created new dataset" not in result.output
        assert "Labeled data has been successfully exported: gantry://datasets" in result.output


def test_export_project_to_dashboard():
    runner = CliRunner()

    app_name = "fake-application"

    cli_args = [
        "--name",
        PROJECT_NAME,
        "--destination",
        app_name,
        "--ls-token",
        LS_TOKEN,
    ]

    with responses.RequestsMock() as resp:
        setup_resp_for_gantry_pings(resp)
        # get project
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/labeling/projects/{PROJECT_NAME}",
            json={
                "data": {"export_storage_id": "fake-export-id", "ls_host": LS_HOST},
                "response": "ok",
            },
        )
        # labelstudio export sync
        resp.add(
            resp.POST,
            f"{LS_HOST}/api/storages/export/s3/fake-export-id/sync",
            json={"response": "ok"},
        )
        # get application
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/models/{app_name}",
            json={"data": {"versions": []}, "response": "ok"},  # just need to return something
        )
        # export labels to dataset
        setup_resp_with_backoff(
            resp,
            resp.POST,
            f"{UNITTEST_HOST}/api/v1/ingest/labeling",
            json={"id": "fake-version-id", "response": "ok"},
            match=[
                matchers.json_params_matcher(
                    {"labeling_project": PROJECT_NAME, "application": app_name}
                )
            ],
        )

        result = runner.invoke(dashboard_export, args=cli_args, env=TEST_ENV)

        assert result.exit_code == 0

        assert "Synced labeled data to S3 bucket" in result.output
        assert "Labeled data has been successfully exported:" in result.output
        assert app_name in result.output


@pytest.mark.parametrize(
    "export",
    [dataset_export, dashboard_export],
)
def test_export_project_no_project(export):
    runner = CliRunner()

    cli_args = [
        "--name",
        PROJECT_NAME,
        "--destination",
        "fake-dataset",
        "--ls-token",
        LS_TOKEN,
    ]

    with responses.RequestsMock() as resp:
        setup_resp_for_gantry_pings(resp)
        # get project (there is no project)
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/labeling/projects/{PROJECT_NAME}",
            status=404,
        )

        result = runner.invoke(export, args=cli_args, env=TEST_ENV)

        assert result.exit_code == 1

        assert result.exc_info[0] == GantryRequestException
        assert "A resource was not found" in result.exc_info[1].args[0]


@pytest.mark.parametrize("num_projects", [0, 1, 2])
def test_list_projects(num_projects):
    runner = CliRunner()

    cli_args = []

    projects = [
        {
            "id": "fake-proj-id",
            "org_id": "fake-org-id",
            "name": "fake-proj-name",
            "source": "fake-source",
            "bucket": "fake-bucket",
            "s3_prefix": "fake-folder",
            "ls_host": "fake-ls-host",
            "created_at": "fake-created-at",
            "import_storage_id": "fake-import-id",
            "export_storage_id": "fake-export-id",
            "status": "fake-status",
        }
    ] * num_projects

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/labeling/projects",
            json={"data": projects, "response": "ok"},
        )

        result = runner.invoke(list, args=cli_args, env=TEST_ENV)

        assert result.exit_code == 0

        if num_projects == 0:
            assert "No labeling projects found" in result.output
        else:
            assert len(result.output.split("\n")) == num_projects + 3
            assert "fake-proj-name" in result.output


def test_validate_source_dataset():
    source_urn = "gantry://datasets/fake-dataset/versions/fake-version"
    ctx = mock.Mock()
    ctx.params = {"api_key": "fake-api-key"}
    HOST = "https://app.gantry.io"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/fake-dataset",
            json={
                "response": "ok",
                "data": {
                    "name": "fake-dataset",
                    "id": "fake-dataset-id",
                    "bucket_name": BUCKET_NAME,
                    "aws_region": REGION,
                    "organization_id": "fake-org-id",
                    "s3_prefix": "fake-prefix",
                    "disabled": False,
                },
            },
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/fake-dataset-id/commits",
            json={
                "response": "ok",
                "data": [
                    {
                        "id": "fake-version",
                        "message": "fake-message",
                        "created_at": "fake-created-at",
                        "created_by": "fake-created-by",
                        "is_latest_commit": True,
                    }
                ],
            },
        )

        assert _validate_source(ctx, None, source_urn) == source_urn


@pytest.mark.parametrize(
    "source_urn",
    [
        "gantry://datasets/fake-dataset/versions/doesnt-exist-version",
        "gantry://datasets/doesnt-exist-dataset/versions/fake-version/",
    ],
)
def test_validate_source_dataset_invalid(source_urn):
    ctx = mock.Mock()
    ctx.params = {"api_key": "fake-api-key"}
    HOST = "https://app.gantry.io"

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/doesnt-exist-dataset",
            status=404,
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/fake-dataset",
            json={
                "response": "ok",
                "data": {
                    "name": "fake-dataset",
                    "id": "fake-dataset-id",
                    "bucket_name": BUCKET_NAME,
                    "aws_region": REGION,
                    "organization_id": "fake-org-id",
                    "s3_prefix": "fake-prefix",
                    "disabled": False,
                },
            },
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/fake-dataset-id/commits",
            json={
                "response": "ok",
                "data": [
                    {
                        "id": "fake-version",
                        "message": "fake-message",
                        "created_at": "fake-created-at",
                        "created_by": "fake-created-by",
                        "is_latest_commit": True,
                    }
                ],
            },
        )

        with pytest.raises(click.exceptions.BadParameter):
            _validate_source(ctx, None, source_urn)


def test_validate_source_batch():
    source_urn = "gantry://apps/fake-app/batches/fake-batch-id"
    ctx = mock.Mock()
    ctx.params = {"api_key": "fake-api-key"}
    HOST = "https://app.gantry.io"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/models/fake-app/schemas",
            json={
                "response": "ok",
                "data": {"version": "0", "id": "test-id"},
            },
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/test-id/batches",
            json={
                "response": "ok",
                "data": {"batches": [{"id": "fake-batch-id"}], "metadata": {"total_count": 1}},
            },
        )

        assert _validate_source(ctx, None, source_urn) == source_urn


@pytest.mark.parametrize(
    "source_urn",
    [
        "gantry://apps/fake-app/batches/doesnt-exist",
        "gantry://apps/doesnt-exist/batches/fake-batch-id",
    ],
)
def test_validate_source_batch_invalid(source_urn):
    ctx = mock.Mock()
    ctx.params = {"api_key": "fake-api-key"}
    HOST = "https://app.gantry.io"

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/models/doesnt-exist/schemas",
            status=404,
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/models/fake-app/schemas",
            json={
                "response": "ok",
                "data": {"version": "0", "id": "test-id"},
            },
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/test-id/batches",
            json={
                "response": "ok",
                "data": {"batches": [{"id": "fake-batch-id"}], "metadata": {"total_count": 1}},
            },
        )

        with pytest.raises(click.exceptions.BadParameter):
            _validate_source(ctx, None, source_urn)



import logging
import uuid
from typing import Optional

from typeguard import typechecked

from gantry.api_client import APIClient
from gantry.applications.core import Application
from gantry.applications.llm import CompletionApplication

logger = logging.getLogger(__name__)


class ApplicationClient:
    def __init__(self, api_client: APIClient) -> None:
        self._api_client = api_client

    @typechecked
    def create_application(
        self, application_name: str, application_type: Optional[str] = None
    ) -> Application:
        """
        Create an Application object by name and type. If the application already exists,
        an exception will be raised. If your organization does not allow you to create
        an application of a certain type, there will be an exception raised as well.

        Example usage:

        .. code-block:: python

            import gantry
            application = gantry.create_application("demo", "Custom")

        Args:
            application_name (str): the name of the application.
            application_type (str, optional): the type of the application. Defaults to Custom.
            You have the following options for application type:
                Custom
                Chat
                Completion
        Returns:
            :class:`gantry.Application`: An object representing the application.
        """
        data = {
            "model_name": application_name,
        }
        if application_type:
            data["model_type"] = application_type

        res = self._api_client.request(
            "POST",
            "/api/v1/applications",
            json=data,
            raise_for_status=True,
        )

        return self._get_application(
            name=application_name,
            id=uuid.UUID(res["data"]["id"]),
            type=application_type,
        )

    @typechecked
    def get_application(self, application_name: str) -> Application:
        """
        Get an Application object by name.

        Example usage:

        .. code-block:: python

            import gantry
            application = gantry.get_application("demo")

        Args:
            application_name (str): the name of the application.
        Returns:
            :class:`gantry.Application`: An object representing the application.
        """
        res1 = self._api_client.request(
            "GET",
            "/api/v1/applications/{}".format(application_name),
            raise_for_status=True,
        )
        res2 = self._api_client.request(
            "GET",
            "/api/v1/applications/{}/schemas".format(application_name),
            raise_for_status=True,
        )

        return self._get_application(
            name=res1["data"]["func_name"],
            id=uuid.UUID(res1["data"]["versions"][0]["application_id"]),
            type=res2["data"]["model_type"],
        )

    @typechecked
    def archive_application(self, application_name: str) -> None:
        """
        Archive an Application.

        Example usage:

        .. code-block:: python

            import gantry
            application = gantry.archive_application("demo")

        Args:
            application_name (str): the name of the application.
        Returns:
            None
        """
        model_id = self.get_application(application_name)._id
        self._api_client.request(
            "DELETE",
            "/api/v1/applications/{}".format(model_id),
            raise_for_status=True,
        )

    @typechecked
    def delete_application(self, application_name: str) -> None:
        """
        Hard delete an Application.

        Example usage:

        .. code-block:: python

            import gantry
            application = gantry.delete_application("demo")

        Args:
            application_name (str): the name of the application.
        Returns:
            None
        """
        model_id = self.get_application(application_name)._id
        self._api_client.request(
            "POST",
            "/api/v1/applications/{}/hard_delete".format(model_id),
            raise_for_status=True,
        )

    def _get_application(self, name: str, id: uuid.UUID, type: Optional[str]) -> Application:
        if type == "Completion":
            return CompletionApplication(
                name=name,
                api_client=self._api_client,
                id=id,
            )
        elif type == "Custom":
            return Application(
                name=name,
                api_client=self._api_client,
                id=id,
            )
        else:
            return Application(
                name=name,
                api_client=self._api_client,
                id=id,
            )



import pathlib
import uuid

import pytest
import responses

from gantry.api_client import APIClient
from gantry.dataset.client import GantryDatasetClient
from gantry.exceptions import DatasetNotFoundError, GantryRequestException

from .conftest import BUCKET_NAME, DATASET_ID, DATASET_NAME, HOST


@pytest.fixture
def test_api_client():
    return APIClient(origin=HOST)


@pytest.fixture
def test_client(test_api_client):
    return GantryDatasetClient(
        api_client=test_api_client,
        working_directory=".",
    )


def test_create_dataset(test_client, test_dataset):
    """
    Test create dataset
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/datasets",
            json={
                "response": "ok",
                "data": test_dataset,
            },
            headers={"Content-Type": "application/json"},
        )
        dataset = test_client.create_dataset(DATASET_NAME)
        assert dataset._dataset_id == DATASET_ID
        assert dataset.dataset_name == DATASET_NAME
        assert dataset.workspace == "."


def test_create_dataset_with_model(test_client, test_dataset):
    """
    Test create dataset with model name
    """
    with responses.RequestsMock() as resp:
        MODEL_ID = str(uuid.uuid4())
        MODEL_NAME = "test_model"
        resp.add(
            resp.GET,
            f"https://test-api/api/v1/applications/{MODEL_NAME}/schemas",
            json={
                "response": "ok",
                "data": {"id": MODEL_ID},
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/datasets",
            json={
                "response": "ok",
                "data": test_dataset,
            },
            headers={"Content-Type": "application/json"},
            match=[
                responses.matchers.json_params_matcher(
                    {
                        "name": DATASET_NAME,
                        "model_id": MODEL_ID,
                        "bucket_name": BUCKET_NAME,
                    }
                )
            ],
        )
        dataset = test_client.create_dataset(
            DATASET_NAME, bucket_name=BUCKET_NAME, app_name="test_model"
        )
        assert dataset._dataset_id == DATASET_ID
        assert dataset.dataset_name == DATASET_NAME
        assert dataset.workspace == "."


def test_get_dataset(test_client, test_dataset):
    """
    Test get dataset
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_NAME}",
            json={
                "response": "ok",
                "data": test_dataset,
            },
            headers={"Content-Type": "application/json"},
        )
        dataset = test_client.get_dataset(DATASET_NAME)
        assert dataset._dataset_id == DATASET_ID
        assert dataset.dataset_name == DATASET_NAME
        assert dataset.workspace == "."


def test_get_dataset_in_app(test_client, test_dataset):
    """
    Test get dataset
    """
    application_name = "test_app"
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application_name}/datasets/{DATASET_NAME}",
            json={
                "response": "ok",
                "data": test_dataset,
            },
            headers={"Content-Type": "application/json"},
        )
        dataset = test_client.get_dataset(DATASET_NAME, application_name)
        assert dataset._dataset_id == DATASET_ID
        assert dataset.dataset_name == DATASET_NAME
        assert dataset.workspace == "."


@pytest.mark.parametrize(
    ("status_code", "expected_error"),
    [(404, DatasetNotFoundError), (500, GantryRequestException)],
)
def test_get_dataset_error(test_client, status_code, expected_error):
    """
    Test get dataset error
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_NAME}",
            status=status_code,
            headers={"Content-Type": "application/json"},
        )
        with pytest.raises(expected_error):
            test_client.get_dataset(DATASET_NAME)


def test_list_dataset_versions(test_client, test_dataset, commit_history):
    """
    Test get dataset commits
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_NAME}",
            json={
                "response": "ok",
                "data": test_dataset,
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={
                "response": "ok",
                "data": commit_history,
            },
            headers={"Content-Type": "application/json"},
        )
        commits = test_client.list_dataset_versions(DATASET_NAME)
        assert len(commits) == 2


@pytest.mark.parametrize("include_deleted", [True, False])
def test_list_datasets(test_client, test_dataset, include_deleted):
    """
    List datasets
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets?include_deleted={include_deleted}",
            json={
                "response": "ok",
                "data": [test_dataset],
            },
            headers={"Content-Type": "application/json"},
        )
        datasets = test_client.list_datasets(include_deleted=include_deleted)
        assert len(datasets) == 1


def test_delete_dataset(test_client, test_dataset):
    """
    Delete specific dataset
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_NAME}",
            json={
                "response": "ok",
                "data": test_dataset,
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.DELETE,
            f"{HOST}/api/v1/datasets/{DATASET_ID}",
            json={
                "response": "ok",
            },
            headers={"Content-Type": "application/json"},
        )

        test_client.delete_dataset(DATASET_NAME)


def test_set_working_dir(test_client):
    assert test_client.working_directory == "."
    test_client.set_working_directory("test_dir")
    assert test_client.working_directory == str(pathlib.Path("test_dir").resolve())



from gantry.query.core.dataframe import GantrySeries
from gantry.query.core.utils import same_size


class GantryDistance(object):
    """
    Gantry distribution distance accessor.
    """

    def __init__(self, api_client) -> None:
        self._api_client = api_client

    @same_size("feat1", "feat2")
    def d1(self, feat1: GantrySeries, feat2: GantrySeries) -> float:
        """
        Computes the D1 distance between the input feature's distributions.

        Args:
            feat1 (GantrySeries): feature as a GantrySeries
            feat2 (GantrySeries): feature to compute dist with as a GantrySeries

        Returns: float d1 distance
        """
        return self._diff_query(feat1, feat2, "d1")

    @same_size("feat1", "feat2")
    def dinf(self, feat1: GantrySeries, feat2: GantrySeries) -> float:
        """
        Computes the maximum distance between the input features's distributions.

        Args:
            feat1 (GantrySeries): feature as a GantrySeries
            feat2 (GantrySeries): feature to compute dist with as a GantrySeries
        Returns: float d_inf distance
        """
        return self._diff_query(feat1, feat2, "dinf")

    def ks(self, feat1: GantrySeries, feat2: GantrySeries) -> float:
        """
        Performs the one-sample Kolmogorov-Smirnov test for goodness of fit between the
        input features's distributions.

        Args:
            feat1 (GantrySeries): feature as a GantrySeries
            feat2 (GantrySeries): feature to compute dist with as a GantrySeries
        Returns: Tuple[float ks distance measure]
        """
        return self._diff_query(feat1, feat2, "ks")

    @same_size("feat1", "feat2")
    def kl(self, feat1: GantrySeries, feat2: GantrySeries) -> float:
        """
        Gets the Kullback-Leibler divergence between the input features's distributions.

        Args:
            feat1 (GantrySeries): feature as a GantrySeries
            feat2 (GantrySeries): feature to compute dist with as a GantrySeries
        Returns: Tuple[float kl divergence]
        """
        return self._diff_query(feat1, feat2, "kl")

    def _diff_query(self, feat1: GantrySeries, feat2: GantrySeries, stat: str) -> float:
        data = feat1.query_info.get_base_query_params()
        data["queries"] = {
            "query": {
                "query_type": "diff",
                "stat": stat,
                "base": {
                    "query_type": "feature",
                    "feature": feat1.name,
                    "model_node_id": feat1.query_info.application_node_id,
                    "start_time": feat1.query_info.start_time,
                    "end_time": feat1.query_info.end_time,
                },
                "other": {
                    "query_type": "feature",
                    "feature": feat2.name,
                    "model_node_id": feat2.query_info.application_node_id,
                    "start_time": feat2.query_info.start_time,
                    "end_time": feat2.query_info.end_time,
                },
            }
        }
        response = self._api_client.request(
            "POST", "/api/v1/aggregate/query", json=data, raise_for_status=True
        )

        try:
            return response["data"]["query"][stat]
        except KeyError:
            raise RuntimeError("Invalid response from API server")



from typing import Optional

import click
from click_aliases import ClickAliasedGroup

import gantry
from gantry.api_client import APIClient
from gantry.utils import check_response_status

STORAGE_TYPES = ["S3", "GCS"]


@click.group(cls=ClickAliasedGroup)
def bucket():
    """
    Use this to manage your bucket with gantry.
    """


@bucket.command(
    aliases=["create"],
    help="Register your bucket with Gantry.",
)
@click.option("--name", type=click.STRING, required=True, help="Bucket name")
@click.option(
    "--region",
    type=click.STRING,
    required=True,
    help="Region of your bucket, for GCS buckets you can set this to auto",
)
@click.option(
    "--storage-type", required=True, type=click.Choice(STORAGE_TYPES, case_sensitive=False)
)
@click.option("--secret", default=None, type=click.STRING, help="Secret name")
def create(name: str, region: str, storage_type: str, secret: Optional[str]):
    gantry.init()
    api_client = gantry.get_client().log_store._api_client
    data = {
        "bucket_name": name,
        "region": region,
        "storage_type": storage_type.upper(),
        "secret_name": secret,
    }

    resp = api_client.request(
        "POST", "/api/v1/customer-resources/buckets", json=data, raise_for_status=True
    )

    check_response_status(resp, f"Fail to register bucket {name}")
    click.secho(f"--> Bucket {name} has been registered.")


@bucket.command(
    aliases=["list"],
    help="List buckets.",
)
def list():
    gantry.init()
    api_client = gantry.get_client().log_store._api_client

    resp = api_client.request("GET", "/api/v1/customer-resources/buckets", raise_for_status=True)

    check_response_status(resp, "Fail to list buckets")
    for data in resp["data"]:
        click.secho(data)


@bucket.command(
    aliases=["update-secrete"],
    help="Update bucket secret key.",
)
@click.option("--bucket", type=click.STRING, required=True, help="Bucket name")
@click.option(
    "--storage-type", required=True, type=click.Choice(STORAGE_TYPES, case_sensitive=False)
)
@click.option("--secret-id", type=click.STRING, default=None, help="Secret id")
def update_secret(bucket: str, storage_type: str, secret_id: Optional[str]):
    gantry.init()
    api_client = gantry.get_client().log_store._api_client
    bucket_id = _get_bucket_id(api_client, bucket, storage_type.upper())
    data = {}
    if secret_id:
        data["secret_id"] = secret_id

    resp = api_client.request(
        "PATCH", f"/api/v1/customer-resources/buckets/{bucket_id}", json=data, raise_for_status=True
    )
    check_response_status(resp, f"Fail to update bucket {bucket}")
    click.secho(f"--> Bucket {bucket} has been updated with key {secret_id}.")


def _get_bucket_id(api_client: APIClient, bucket_name: str, storage_type: str) -> str:
    resp = api_client.request(
        "GET",
        f"/api/v1/customer-resources/buckets/{bucket_name}?storage_type={storage_type.upper()}",
        raise_for_status=True,
    )
    return resp["data"]["id"]



import functools
import warnings
from typing import Dict, List, Optional

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

import pandas as pd

from gantry.query.core.dataframe import GantrySeries
from gantry.query.core.utils import same_size


def _prepare_pandas_dataframe(stat, results, dropna) -> pd.DataFrame:
    df = pd.DataFrame.from_records(results)
    if df.empty:
        return df

    # note that if num_points is 1, the timestamp is start_time
    # this is probably better than using end_time because
    # end_time is possibly in the future
    df.time_stamp = pd.to_datetime(df.time_stamp)
    df.rename(columns={"value": stat, "time_stamp": "timestamp"}, inplace=True)

    if dropna:
        df = df.dropna()

    return df.set_index("timestamp")


def categorical(func):
    @functools.wraps(func)
    def categorical_dec(self, outputs: GantrySeries, feedback: GantrySeries, *args, **kwargs):
        if outputs.datatype != "float":
            return func(self, outputs, feedback, *args, **kwargs)
        else:
            raise NotImplementedError()

    return categorical_dec


def regression(func):
    @functools.wraps(func)
    def regression_dec(self, outputs: GantrySeries, feedback: GantrySeries, *args, **kwargs):
        if outputs.datatype == "float":
            return func(self, outputs, feedback, *args, **kwargs)
        else:
            raise NotImplementedError()

    return regression_dec


def score(func):
    @functools.wraps(func)
    def score_dec(self, outputs: GantrySeries, feedback: GantrySeries, *args, **kwargs):
        if outputs.datatype == "float" and feedback.datatype != "float":
            return func(self, outputs, feedback, *args, **kwargs)
        else:
            raise NotImplementedError(
                "Outputs and feedback do not have the appropriate types to calculate score metrics"
            )

    return score_dec


def bool_type_check(func):
    @functools.wraps(func)
    def bool_dec(self, data_node: GantrySeries, *args, **kwargs):
        if data_node.datatype == "bool":
            return func(self, data_node, *args, **kwargs)
        else:
            raise NotImplementedError("Series type is not bool!")

    return bool_dec


def check_comparable(series_a, series_b):
    if not all(
        [filter_a == filter_b for filter_a, filter_b in zip(series_a.filters, series_b.filters)]
    ):
        raise ValueError(
            "Series A and series B do not have the same filters. "
            "Series A filters: {}\nSeries B filters: {}".format(series_a.filters, series_b.filters)
        )


class GantryMetric(object):
    """
    Gantry metric computations.

    Inspired by and references `sklearn metrics
    <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_

    For all queries, we use the first available GantrySeries's QueryInfo to obtain
    the relevant query metadata window.
    """

    def __init__(self, api_client) -> None:
        self._api_client = api_client

    def _metric_query(
        self,
        data_nodes: List[GantrySeries],
        stat: str,
        num_points: Optional[int] = None,
        stat_kwargs: Optional[dict] = None,
    ) -> List[Dict]:
        if (num_points is not None) and (num_points < 1):
            raise ValueError(
                "num_points must be greater than 0 if passed. Got {}.".format(num_points)
            )

        if not (1 <= len(data_nodes) <= 2):
            raise ValueError("Please provide 1 or 2 data_nodes in your query.")

        # TODO if data_nodes don't have the same parent dataframe
        # or have different query_info, we should raise an error. Can check filters.
        if len(data_nodes) == 2:
            check_comparable(data_nodes[0], data_nodes[1])

        data = {
            "start_time": data_nodes[0].query_info.start_time,
            "end_time": data_nodes[0].query_info.end_time,
            "num_points": num_points,
            "queries": {
                "query_label": {
                    "query_type": "data_node_ids",
                    "model_node_id": data_nodes[0].query_info.application_node_id,
                    "stat": stat,
                    "stat_kwargs": stat_kwargs or {},
                    "filters": data_nodes[0].filters,
                    "start_time": data_nodes[0].query_info.start_time,
                    "end_time": data_nodes[0].query_info.end_time,
                    "data_node_ids": [data_node.id for data_node in data_nodes],
                },
            },
        }

        response = self._api_client.request(
            "POST", "/api/v1/time_series/query", json=data, raise_for_status=True
        )
        return response["data"]["query_label"].get("points", [])

    @categorical
    @same_size("outputs", "feedback")
    def accuracy_score(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Categorical metric - accuracy

        In multilabel classification, this computes the set of |outputs| which exactly match
        the available |feedback|.

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as
                a GantrySeries
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) accuracy score
        """
        stat = "accuracy_score"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    @regression
    @same_size("outputs", "feedback")
    def mean_squared_error(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
        multioutput: Literal["uniform_average", "raw_values"] = "uniform_average",
        squared: bool = True,
    ) -> pd.DataFrame:
        """
        Regression metric- mean squared error

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as a GantrySeries
            dropna (bool, defaults to False): if True, drop rows with NaN values in result
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into.
            multioutput ("uniform_average" or "raw_values"): type of
                averaging to use when computing the metric. Defaults to
                "uniform_average".
            squared (boo, defaults to True): if True, return the squared error.

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) mean_squared_error
        """
        stat = "mean_squared_error"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
            stat_kwargs={"multioutput": multioutput, "squared": squared},
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    @categorical
    @same_size("outputs", "feedback")
    def confusion_matrix(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Categorical metric - confusion matrix
        The confusion matrix is a matrix :math:`C` where :math:`C_{i, j}`
        represents the number of times a data point from the class :math:`i`
        was predicted to be in class :math:`j`.

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as a
                GantrySeries.
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1)
                confusion_matrix
        """
        stat = "confusion_matrix"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    @categorical
    @same_size("outputs", "feedback")
    def f1_score(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
        average: Literal["micro"] = "micro",
    ) -> pd.DataFrame:
        """
        Categorical metric - F1 score

        It is computed as the harmonic mean of precision and recall:
        F1 = 2 * (precision * recall) / (precision + recall)
        In multiclass classification, this is the average of the F1 score for all available classes.

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as a
                GantrySeries.
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            average ("micro"): type of averaging to use when computing
                the metric. Currently only "micro" supported, which is
                the default value.

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) f1_score
        """
        stat = "f1_score"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
            stat_kwargs={"average": average},
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    @regression
    @same_size("outputs", "feedback")
    def r2_score(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
        multioutput: Literal[
            "uniform_average", "raw_values", "variance_weighted"
        ] = "uniform_average",
    ) -> float:
        """
        Regression metric- R^2 coefficient of determination

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as a
                GantrySeries.
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            multioutput ("uniform_average", "raw_values", "variance_weighted"):
                type of averaging to use when computing the metric. Defaults
                to "uniform_average".

        Returns:
            (float) R^2 score.
        """
        stat = "r2_score"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
            stat_kwargs={"multioutput": multioutput},
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    @categorical
    @same_size("outputs", "feedback")
    def precision_score(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
        average: Literal["micro"] = "micro",
    ) -> pd.DataFrame:
        """
        Categorical metric - precision score

        precision =
        (number of true positives) / ((number of true positives) + (number of false positives))

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as a
                GantrySeries.
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            average ("micro"): type of averaging to use when computing
                the metric. Currently only "micro" supported, which is
                the default value.

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) precision_score
        """
        stat = "precision_score"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
            stat_kwargs={"average": average},
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    @categorical
    @same_size("outputs", "feedback")
    def recall_score(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
        average: Literal["micro"] = "micro",
    ) -> pd.DataFrame:
        """
        Categorical metric - recall score

        recall =
        (number of true positives) / ((number of true positives) + (number of false negatives))

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as a
                GantrySeries.
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            average ("micro"): type of averaging to use when computing
                the metric. Currently only "micro" supported, which is
                the default value.

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) recall_score
        """
        stat = "recall_score"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
            stat_kwargs={"average": average},
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    @score
    @same_size("outputs", "feedback")
    def roc_auc_score(
        self,
        outputs: GantrySeries,
        feedback: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Classification score metric - the area under the ROC curve

        Args:
            outputs (GantrySeries): predictions as a GantrySeries
            feedback (GantrySeries): labels to compare against as a
                GantrySeries.
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) roc_auc_score
        """
        stat = "roc_auc_score"
        results = self._metric_query(
            [outputs, feedback],
            stat,
            num_points=num_points,
        )
        return _prepare_pandas_dataframe(stat, results, dropna=dropna)

    def _percent_query(
        self,
        stat,
        data_node: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        if dropna:
            data_node = data_node[data_node.notna()]  # type: ignore

        results = self._metric_query(
            [data_node],
            stat,
            num_points=num_points,
        )
        return _prepare_pandas_dataframe(stat, results, dropna=False)

    def percent_null(
        self,
        data_node: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Percent null/NaN

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) percent_null
        """
        if dropna:
            warnings.warn(
                "'dropna' parameter is no longer valid in this "
                "query as it was a bug. Please remove it as it "
                "will be removed in future versions"
            )

        stat = "percent_null"
        return self._percent_query(stat, data_node, False, num_points)

    @bool_type_check
    def percent_true(
        self,
        data_node: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Percent true

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) percent_true
        """
        stat = "percent_true"
        return self._percent_query(stat, data_node, dropna, num_points)

    @bool_type_check
    def percent_false(
        self,
        data_node: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Percent false

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) percent_false
        """
        stat = "percent_false"
        return self._percent_query(stat, data_node, dropna, num_points)

    @bool_type_check
    def percent_true_not_null(
        self,
        data_node: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Percent true after excluding null values

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) percent_true_not_null
        """
        stat = "percent_true_not_null"
        return self._percent_query(stat, data_node, dropna, num_points)

    @bool_type_check
    def percent_false_not_null(
        self,
        data_node: GantrySeries,
        dropna: bool = False,
        num_points: int = 1,
    ) -> pd.DataFrame:
        """
        Percent false after excluding null values

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            (pd.DataFrame) pd.DataFrame of shape (num_points, 1) percent_false_not_null
        """
        stat = "percent_false_not_null"
        return self._percent_query(stat, data_node, dropna, num_points)



import datetime
import json
import logging
import uuid
from functools import wraps
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from gantry import dataset
from gantry.api_client import APIClient
from gantry.automations.automations import Automation
from gantry.exceptions import ClientNotInitialized, GantryRequestException
from gantry.logger.constants import BatchType
from gantry.logger.main import _upload_data_as_batch, log, log_file
from gantry.query.core.dataframe import GantryDataFrame
from gantry.query.core.utils import get_application_node_id
from gantry.query.time_window import RelativeTimeWindow, TimeWindow

logger = logging.getLogger(__name__)


class Run:
    """
    Context manager for logging.
    """

    # Class variable to store the current run instance
    _current_instance = None

    def __init__(self, application, tags: Optional[Dict], name: str):
        if not isinstance(application, Application):
            raise TypeError("Your input must be an Application.")
        self.run_id = str(uuid.uuid4())

        self.name: str = name
        self.application: Application = application
        self.tags: Dict = tags or {}
        self.prediction_and_feedback_events: List[Dict] = []
        self.prediction_events: List[Dict] = []
        self.feedback_events: List[Dict] = []
        self.join_keys: List[str] = []

    def __enter__(self):
        logger.info("Starting run...")
        logger.info("Run ID: " + str(self.run_id))
        Run._current_instance = self
        return self

    def add_prediction_and_feedback_events(self, events: List[Dict]):
        self.prediction_and_feedback_events.extend(events)

    def add_prediction_events(self, events: List[Dict]):
        self.prediction_events.extend(events)

    def add_feedback_events(self, events: List[Dict]):
        self.feedback_events.extend(events)

    def prepare_join_key_list(self):
        for event in self.prediction_and_feedback_events:
            if event["feedback_id"] not in self.join_keys:
                self.join_keys.append(event["feedback_id"])
        for event in self.feedback_events:
            if event["feedback_id"] not in self.join_keys:
                self.join_keys.append(event["feedback_id"])

    def merge_events(self):
        """
        Merge prediction and feedback events into a single list of events.
        """
        merge_join_keys = {}
        for event in self.prediction_events:
            if event["feedback_id"] not in merge_join_keys:
                merge_join_keys[event["feedback_id"]] = {}
            merge_join_keys[event["feedback_id"]]["prediction"] = event

        for event in self.feedback_events:
            if event["feedback_id"] in merge_join_keys:
                merge_join_keys[event["feedback_id"]]["feedback"] = event

        for key, value in merge_join_keys.items():
            if key not in self.join_keys:
                self.join_keys.append(key)
            event = {}
            if "feedback" not in value:
                event.update(value["prediction"])
                event.update({"feedback": {"ground_truth_label": None}})
            elif "prediction" not in value:
                continue
            else:
                event.update(value["prediction"])
                event["metadata"].update(value["feedback"].pop("metadata"))
                event.update(value["feedback"])
            self.prediction_and_feedback_events.append(event)

    def __exit__(self, exc_type, exc_val, exc_tb):
        logging.info("Ending run...")
        self.prepare_join_key_list()
        self.merge_events()

        if len(self.prediction_and_feedback_events) - len(self.prediction_events) > 0:
            _upload_data_as_batch(
                self.application._name,
                version=None,
                events=self.prediction_and_feedback_events,
                batch_type=BatchType.RECORD,
                run_id=self.run_id,
                run_name=self.name,
                run_tags=self.tags,
            )
        elif len(self.prediction_events) > 0:
            _upload_data_as_batch(
                self.application._name,
                version=None,
                events=self.prediction_events,
                batch_type=BatchType.PREDICTION,
                run_id=self.run_id,
                run_name=self.name,
                run_tags=self.tags,
            )
        elif len(self.feedback_events) > 0:
            _upload_data_as_batch(
                self.application._name,
                version=None,
                events=self.feedback_events,
                batch_type=BatchType.FEEDBACK,
                run_id=self.run_id,
                run_name=self.name,
                run_tags=self.tags,
            )
        Run._current_instance = None

    def end(self):
        self.__exit__()


def run_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if Run._current_instance:
            if "inputs" in result[0] and "feedback" in result[0]:
                Run._current_instance.add_prediction_and_feedback_events(result)
            elif "inputs" in result[0]:
                Run._current_instance.add_prediction_events(result)
            elif "feedback" in result[0]:
                Run._current_instance.add_feedback_events(result)
        return result

    return wrapper


class Application:
    """
    A class to represent a single application. This class is not directly initialized,
    but is instead initialized by calling `gantry.create_application("my-app")`, or
    `gantry.get_application("my-app")`.
    """

    def __init__(
        self,
        name: str,
        api_client: APIClient,
        id: Optional[uuid.UUID] = None,
        organization_id: Optional[uuid.UUID] = None,
    ):
        """
        Note: Please use gantry.create_application() to create an application instead
        of defining an application directly by calling this method.
        """
        self._api_client = api_client
        if self._api_client is None:
            raise ClientNotInitialized()
        self._id = id or uuid.uuid4()
        self._name = name
        self._organization_id = organization_id
        self._version = 1

    def log_file(
        self,
        filepath: str,
        version: Optional[Union[int, str]] = None,
        timestamp: Optional[str] = None,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        feedbacks: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        feedback_id: Optional[str] = None,
        feedback_keys: Optional[List[str]] = None,
        sep: str = ",",
        join_key: Optional[str] = None,
    ):
        """
        Ingest a file containing records into Gantry. File MUST contain a header
        line with column names.

        This method provides parameters to customize column name
        mappings to inputs/outputs/feedback/timestamps/etc.

        All specifications are specifications of file column names

        IMPORTANT: SDK validation will be done only at the header level
        (validating inputs/outputs/feedback parametrization is valid).
        Data itself won't be validated due to potential limitations with
        file size.

        Args:
            filepath (str): path to the file.
            version (optional, Union[int, str]): Version of the function schema to use
                for validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            timestamp (optional, str): by default, timestamp values will be fetched
                from a 'timestamp' named column. Set this parameter to reference a
                different column.
            inputs (optional, list[str]): by default, all columns that start with 'input'
                will be considered inputs. Alternatively, you can provide a list of
                column names using this parameter to be considered inputs.
            outputs (optional, list[str]): by default, all columns that start with 'output'
                will be considered outputs. Alternatively, you can provide a list of
                column names using this parameter to be considered outputs.
            feedbacks (optional, list[str]): by default, all columns that start with 'feedback'
                will be considered feedbacks. Alternatively, you can provide a list of
                column names using this parameter to be considered feedbacks.
            tags (optional, list[str]): by default, all columns that start with 'tags'
                will be considered tags. Alternatively, you can provide a list of
                column names using this parameter to be considered tags.
            feedback_id (optional, str): by default, feedback_id values will be fetched
                from a 'feedback_id' named column. Set this parameter to reference
                another column.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_key' instead.
            feedback_keys (optional, list[str]): provide a list of column names to be used
                as feedback_keys.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_key' instead.
            sep (str, defaults to ','): Character separator, defaults to ',' (for CSV files).
            join_key (optional, str): by default, feedback_id values will be fetched
                from a 'join_key' named column. Set this parameter to reference
                another column.
        """
        log_file(
            self._name,
            filepath,
            version=version,
            timestamp=timestamp,
            inputs=inputs,
            outputs=outputs,
            feedbacks=feedbacks,
            tags=tags,
            feedback_id=feedback_id,
            feedback_keys=feedback_keys,
            sep=sep,
            join_key=join_key,
        )

    @run_decorator
    def log(
        self,
        inputs: Optional[Union[dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        outputs: Optional[
            Union[Any, List[Any], dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]
        ] = None,
        feedbacks: Optional[Union[dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[
            Union[datetime.datetime, List[datetime.datetime], pd.DatetimeIndex, np.ndarray]
        ] = None,
        sort_on_timestamp: bool = True,
        sample_rate: float = 1.0,
        as_batch: Optional[bool] = False,
        version: Optional[Union[int, str]] = None,
        tags: Optional[Union[Dict[str, str], List[Dict[str, str]]]] = None,
        row_tags: Optional[
            Union[Dict[str, str], List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]
        ] = None,
        global_tags: Optional[Dict[str, str]] = None,
        join_keys: Optional[Union[str, List[str], pd.Series]] = None,
    ):
        """
        Ingests an event or a batch of events containing predictions (inputs and outputs), feedback, or both simultaneously.

        Example:

        .. code-block:: python

            app.log(
                inputs=[{'A': 100}, {'A': 101}],
                outputs=[{'B': 'foo'}, {'B': 'bar'}],
                version=1,
                feedbacks=[{'B': 'bar'}, {'B': 'foo'}],
                global_tags={"env": "environment1"},
                join_keys=["12345", "67890"]
            )

        Args:
            inputs (optional, Union[dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]) : A list of prediction inputs. `inputs[i]`
                is a dict of the features for the i-th prediction to be logged.
            outputs (optional, Union[Any, List[Any], dict, List[dict], pd.DataFrame, pd.Series, np.ndarray): A list of prediction outputs. `outputs[i]`
                should be the application output for the prediction with features `inputs[i]`.
            feedbacks (optional, Union[dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]): A list of feedbacks. `feedbacks[i]`
                is a dict of the features for the i-th prediction to be logged.
            ignore_inputs (optional, List[str]): A list of names of input features that should not
                be monitored.
            timestamps (optional, Union[List[datetime.datetime], pd.DatetimeIndex, np.array[datetime.datetime]):  # noqa: E501
                A list of prediction timestamps.
                If specified, `timestamps[i]` should be the timestamps for the i-th prediction.
                If timestamps = None (default), then the prediction
                timestamp defaults to the time when `log_records` is called.
            sort_on_timestamp (bool, defaults to True): Works when timestamps are provided.
                Sort using the given timestamp. Default to True.
            sample_rate: Used for down-sampling. The probability as a float that each record
                will be sent to Gantry.
            as_batch (bool, defaults to False): Whether to add batch metadata and tracking
                in the 'batch' section of the dashboard
            version (optional, Union[int, str]): Version of the function schema to use for
                validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            tags (optional, Optional[Union[Dict[str, str], List[Dict[str, str]]]): A tag is a
                label that you assign to your data. E.g. you can specify which environment
                the data belongs to by setting "env" tag like this tags = {"env": "environment1"}
                if not assigned we will use Gantry client's environment value as the defaults
                environment. If this parameter is a dict, tags will apply to all records.
                Alternatively, you can pass a list of dicts to apply tags to each record
                independantly.
                IMPORTANT: this parameter is in deprecation mode and it will be removed soon.
                Use row_tags and global_tags instead.
            row_tags (optional, Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]]): Specify row level tags.
                If provided, this parameter should be a list of tags to be applied to each of the records.
                row_tags[i] will contain a dictionary of strings containing the tags to attach to the
                i-th record. Alternatively, tags may be specified by passing in a DataFrame, Series,
                or Array, like inputs.

                For batch global tags, use the 'global_tags' parameter instead.
            global_tags (optional, Dict[str, str]): Specify a set of tags that will be attached to
                all ingested records from this batch. For record specific tags, you can use
                'row_tags' param. Only used when log is not called within a run.
            join_keys  (optional, Union[List[str], pd.Series[str]]): provide keys to identify
                each record. These keys can be used later to provide feedback. If not provided,
                a random record key will be generated for each record.
        """
        run_id = None
        run_tags = None
        if Run._current_instance:
            run_id = Run._current_instance.run_id
            run_tags = Run._current_instance.tags
        return log(
            application=self._name,
            inputs=inputs,
            outputs=outputs,
            feedbacks=feedbacks,
            ignore_inputs=ignore_inputs,
            tags=tags,
            timestamps=timestamps,
            sort_on_timestamp=sort_on_timestamp,
            sample_rate=sample_rate,
            row_tags=row_tags,
            global_tags=global_tags,
            join_keys=join_keys,
            as_batch=as_batch,
            run_id=run_id,
            run_tags=run_tags,
            version=version,
        )

    def query(
        self,
        time_window: Optional[Union[TimeWindow, RelativeTimeWindow]] = None,
        version: Optional[Union[int, str]] = None,
        env: Optional[str] = None,
        filters: Optional[List[dict]] = None,
        tags: Optional[dict] = None,
    ):
        """
        Query for a window of data for this application, with given parameters.

        Example:

        .. code-block:: python

            time_window = RelativeTimeWindow(window_length = datetime.timedelta(days=1),
                offset = datetime.timedelta(minutes=1))
            query = app.query(
                time_window,
                tags = {"champion": "messi"},
                filters=[
                    {
                        "feature_name": "inputs.speed",
                        "lower_bound": 0,
                        "upper_bound": 3
                    }
                ]
            )

        Args:
            time_window (optional, Union[TimeWindow, RelativeTimeWindow]): A time window object.
            version (optional, Union[int, str]): The version of the application to query.
            env (optional, str) : The environment of the application to query.
            filters (optional, list[dict]) : A list of filters to apply to the query.
            tags (optional, dict) : A dictionary of tags to apply to the query.

        Returns:
            Gantry Dataframe with all the query information to fetch data.
        """
        start_time: Union[str, datetime.datetime] = "-24H"
        end_time: Union[str, datetime.datetime] = "now"
        if time_window:
            start_time = time_window.start_time
            end_time = time_window.end_time

        return GantryDataFrame(
            api_client=self._api_client,
            application=self._name,
            start_time=start_time,
            end_time=end_time,
            version=version,
            env=env,
            filters=filters,
            tags=tags,
        )

    def save_query(self, name: str, query: GantryDataFrame):
        """
        Save a query to the Gantry database.

        Example:

        .. code-block:: python

            query = app.query(....)
            app.save_query("demo_query", query)

        Args:
            name (str): The name of the query.
            query (GantryDataFrame): The query to save.
        """
        data = {
            "name": name,
            "env": query.env,
            "tags": query.tags or {},
            "filters": query.filters or [],
            "details": {"color": "default"},
        }
        if query.start_time or query.end_time:
            if not (query.start_time and query.end_time):
                raise ValueError("Both start_time and end_time must be specified.")
            data["start_time"] = query.start_time  # type: ignore
            data["end_time"] = query.end_time  # type: ignore

        try:
            model_node_id = get_application_node_id(
                self._api_client, self._name, version=query.version
            )
            data["model_node_id"] = model_node_id
        except GantryRequestException:
            raise ValueError(
                f"Application {self._name}[{query.version if query.version else 'latest'}]"
                + "does not exist."
            )
        if name in [query["name"] for query in self.list_queries()]:
            self._api_client.request("PATCH", "/api/v1/queries", json=data, raise_for_status=True)
            logger.info("Query updated successfully.")
        else:
            self._api_client.request("POST", "/api/v1/queries", json=data, raise_for_status=True)
            logger.info("Query saved successfully.")

    def get_query(self, name: str) -> GantryDataFrame:
        """
        Get a saved query from the Gantry database.

        Example:

        .. code-block:: python

            query = app.get_query("demo_query")

        Args:
            name (str) : The name of the query.
        Returns:
            Gantry Dataframe with all the query information to fetch data.
        """
        try:
            model_node_id = get_application_node_id(self._api_client, self._name)
            query = self._api_client.request(
                "GET", f"/api/v1/applications/{model_node_id}/queries/{name}"
            )
            query = query["data"]
            deserialized_tags = {}
            for tag in query["tags"]:
                deserialized_tags[tag["tag_name"]] = tag["tag_value"]

            return GantryDataFrame(
                api_client=self._api_client,
                application=self._name,
                start_time=query["start_time"],
                end_time=query["end_time"],
                filters=query["filters"],
                tags=deserialized_tags,
            )
        except GantryRequestException:
            raise ValueError("Application does not exist.")

    def list_queries(self):
        """
        List all saved queries for this application.
        Example:

        .. code-block:: python

            queries = app.list_queries()
            for query in queries:
                print(query["name"])

        Returns:
            A list of queries (in Python dict).
        """
        query = self._api_client.request("GET", f"/api/v1/applications/{self._id}/queries")
        return query["data"]

    def start_run(self, name: str, tags: Optional[dict] = None):
        """
        Start a run for this application.

        Example:

        .. code-block:: python

            with app.start_run(name="Demo_run", tags = {"champion": "messi"}) as run:
                app.log(inputs=inputs, outputs=outputs, feedbacks=feedback, join_keys=join_keys)
                app.log(inputs={...}, outputs={...}, join_keys='12345')
                app.log(....)

        Args:
            name (str): Name of the run.
            tags (optional, dict): A dictionary of tags to apply to the run.

        Returns:
            A Run object.
        """
        return Run(self, name=name, tags=tags)

    def get_run(self, name: Optional[str] = None, tags: Optional[dict] = None):
        """
        Get a run in this application, filtered by tags or name.

        Example:

        .. code-block:: python

            job_info = app.get_run(name="Demo_run_prediction")

        .. code-block:: python

            jobs_info = app.get_run(tags={"champion": "argentina"})

        Args:
            name (optional, str): Name of the run.
            tags (optional, dict): Tags of the run.

        Returns:
            An ingestion job object associated with the requested run.
        """
        if name:
            response = self._api_client.request(
                "GET",
                f"/api/v1/applications/{self._id}/jobs",
                params={"run_name": name},
                raise_for_status=True,
            )
            return response["data"]["jobs"]
        if tags:
            response = self._api_client.request(
                "GET",
                f"/api/v1/applications/{self._id}/jobs",
                params={"run_tags": json.dumps(tags)},
                raise_for_status=True,
            )
            return response["data"]["jobs"]

    def add_automation(self, automation: Automation):
        """
        Add an automation to this application.

        Example:

        .. code-block:: python

            automation = Automation(...)
            app.add_automation(automation)

        Args:
            automation (Automation): An Automation object.

        Returns:
            None
        """
        automation.add_to_application(self._name)

    def get_automation(self, automation_name: Optional[str] = None):
        """
        Get an automation by name

        Example:

        .. code-block:: python

            automation = app.get_automation(automation_name="demo_automation")

        Args:
            automation_name (optional, str): Name of the automation.
        Returns:
            An Automation object.
        """
        resp = self._api_client.request(
            "GET", f"/api/v1/automations/{automation_name}", params={"application_name": self._name}
        )
        data = resp["data"]
        data["application"] = self.get_schema()["func_name"]
        return Automation.from_dict(data)

    def list_automations(self):
        """
        List all automations for this application.

        Example:

        .. code-block:: python

            automations = app.list_automations()
            for automation in automations:
                print(automation["name"])

        Returns:
            A list of Automation objects.
        """
        resp = self._api_client.request(
            "GET", "/api/v1/automations", params={"application_name": self._name}
        )
        return resp["data"]

    def list_workspaces(self):
        """
        Get all workspaces for this application.

        Example:

        .. code-block:: python

            workspaces = app.list_workspaces()
            for workspace in workspaces:
                print(workspace["name"])

        Returns:
            A list of Workspace objects.
        """
        res = self._api_client.request(
            "GET", "/api/v1/workspaces", params={"application_id": self._id}
        )
        if "workspaces" in res["data"]:
            return res["data"]["workspaces"]
        return res["data"]

    def get_schema(self):
        """
        Get the schema for this application.

        Example:

        .. code-block:: python

            schema = app.get_schema()
            print(schema["func_name"]) -> application name

        Returns:
            The application schema (in Python dict).
        """
        # GET /api/v1/applications/<model_name>/schemas
        res = self._api_client.request("GET", f"/api/v1/applications/{self._name}/schemas")
        return res["data"]

    def create_dataset(self, name: str, bucket_name: Optional[str] = None):
        """
        Create a dataset for this application.

        Example:

        .. code-block:: python

            dataset = app.create_dataset(name="demo_dataset", bucket_name="demo_bucket")

        Args:
            name (str): The name of the dataset.
            bucket_name (optional, str): The bucket name of the dataset.

        Return:
            :class:`gantry.dataset.GantryDataset`: an object representing the created dataset.
        """
        return dataset.create_dataset(name=name, bucket_name=bucket_name, app_name=self._name)

    def list_datasets(self, include_deleted: bool = False) -> List[Dict[str, Any]]:
        """
        List all datasets for this application.

        Example:

        .. code-block:: python

            datasets = app.list_datasets()
            for dataset in datasets:
                print(dataset["name"])

        Args:
            include_deleted (optional, bool): Will include deleted datasets if set to `true`
                and will omit them otherwise.
        Returns:
            ``List[Dict[str, Any]]``: List of dictionaries, each representing one
            dataset and associated metadata.
        """
        return dataset.list_datasets(include_deleted=include_deleted, model_node_id=self._id)

    # NotImplemented: chat/completion functions

    def create_version(self, *args, **kwargs):
        """
        :meta private:
        """
        raise NotImplementedError(
            "Versions are not supported for this application.\n"
            "Versions are only supported for completion applications."
        )

    def get_version(self, *args, **kwargs):
        """
        :meta private:
        """
        raise NotImplementedError(
            "Versions are not supported for this application.\n"
            "Versions are only supported for completion applications."
        )

    def log_completion(self, *args, **kwargs):
        """
        :meta private:
        """
        raise NotImplementedError("Completions cannot be logged for a non-completion application.")



import functools
import inspect
import logging
import os
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

from gantry.api_client import APIClient
from gantry.const import PROD_API_URL
from gantry.dataset.client import GantryDatasetClient
from gantry.exceptions import ClientNotInitialized

_DATASET: Optional[GantryDatasetClient] = None
logger_obj = logging.getLogger(__name__)


def _dataset_client_alias(f):
    doc = "Alias for :meth:`gantry.dataset.client.GantryDatasetClient.{}`".format(f.__name__)
    signature = inspect.signature(getattr(GantryDatasetClient, f.__name__))
    orig_doc = inspect.getdoc(getattr(GantryDatasetClient, f.__name__))

    if orig_doc:
        doc += "\n\n{}".format(orig_doc)
    signature = signature.replace(parameters=tuple(signature.parameters.values())[1:])

    f.__doc__ = doc
    f.__signature__ = signature

    # This decorator also checks that the _CLIENT
    # has been initialized
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if _DATASET is None:
            raise ClientNotInitialized()

        return f(*args, **kwargs)

    return wrapper


def _init(
    api_key: Optional[str] = None,
):
    """
    Initialize the dataset functionality. Initialization should happen before any dataset operation.

    Example:

    .. code-block:: python

       import gantry.dataset as gdataset

       gdataset.init(api_key="foobar")

    Args:
        api_key (str): The API key. Users can fetch the API key from the dashboard.
    """
    backend = os.environ.get("GANTRY_LOGS_LOCATION", PROD_API_URL)
    working_directory = os.environ.get("GANTRY_DATASET_WORKING_DIR", str(Path().absolute()))
    parsed_origin = urlparse(backend)
    if parsed_origin.scheme not in ("http", "https"):
        raise ValueError(
            "Invalid backend. http or https backends " + "supported. Got {}".format(backend)
        )

    # Check environment if api_key is not provided
    api_key = os.environ.get("GANTRY_API_KEY") if api_key is None else api_key

    if not api_key:
        raise ValueError(
            """
            No API key provided. Please pass the api_key parameter or set the GANTRY_API_KEY
            environment variable.
            """
        )

    global _DATASET

    api_client = APIClient(backend, api_key)
    _DATASET = GantryDatasetClient(
        api_client=api_client,
        working_directory=working_directory,
    )  # type: ignore[union-attr]

    if not _DATASET.ping():
        logger_obj.warning(
            "Gantry services not reachable. Check provided URL "
            f"[{backend}] is pointing to the correct address"
        )
        return

    if not _DATASET.ready():
        logger_obj.warning("Gantry services won't receive traffic. Check if API Key is valid")

    # TODO:// only show this warning when customer use dataset sdk
    # if not os.environ.get("GANTRY_DATASET_WORKING_DIR"):
    #     logger_obj.warning(
    #         "You haven't set dataset working directory yet, by default Gantry will "
    #         + f"use {str(Path().absolute())}. You can overwrite the default setting"
    #         + "using set_working_directory method."
    #     )


@_dataset_client_alias
def set_working_directory(*args, **kwargs):
    _DATASET.set_working_directory(*args, **kwargs)


@_dataset_client_alias
def create_dataset(*args, **kwargs):
    return _DATASET.create_dataset(*args, **kwargs)


@_dataset_client_alias
def get_dataset(*args, **kwargs):
    return _DATASET.get_dataset(*args, **kwargs)


@_dataset_client_alias
def list_dataset_versions(*args, **kwargs):
    return _DATASET.list_dataset_versions(*args, **kwargs)


@_dataset_client_alias
def list_datasets(*args, **kwargs):
    return _DATASET.list_datasets(*args, **kwargs)


@_dataset_client_alias
def delete_dataset(*args, **kwargs):
    return _DATASET.delete_dataset(*args, **kwargs)



import atexit
import json
import logging
import logging.handlers
from dataclasses import asdict
from queue import Queue
from typing import Callable, Collection, Iterator, List, Optional, cast

import backoff
import requests

from gantry.api_client import APIClient
from gantry.exceptions import GantryRequestException
from gantry.logger.consumer import BatchConsumer
from gantry.logger.types import (
    IngestFromDataConnectorRequest,
    IngestFromDataConnectorResponse,
)
from gantry.serializers import EventEncoder

logger = logging.getLogger(__name__)


BATCH_UPLOAD_MAX_TRIES = 3
BACKOFF_INTERVAL_FOR_ERROR_CODE_SECS = 5
BACKOFF_INTERVAL_FOR_TIMEOUT_SECS = 1


def non_retry_error(e: Exception) -> bool:
    gre = cast(GantryRequestException, e)
    return gre.status_code != 502


def backoff_retry_handler(backoff_ctx: backoff._typing.Details):
    logger.warning("Experiencing timeouts. Please check your connection.")


def backoff_giveup_handler(backoff_ctx: backoff._typing.Details):
    logger.error(
        f"""Failed to log events due to timeout. Have tried {backoff_ctx['tries']}
         times for {backoff_ctx['elapsed']} seconds. Please check your connection."""
    )


def _batch_iterator_factory(
    collection: Collection, batch_size: int
) -> Callable[[], Iterator[List]]:
    if batch_size <= 0:
        raise ValueError(f"Batch size needs to be a positive int, not {batch_size}")

    def _iterator():
        size = len(collection)
        for ndx in range(0, size, batch_size):
            yield collection[ndx : min(ndx + batch_size, size)]  # noqa: E203

    return _iterator


class BaseLogStore:
    def log(self, application: str, event: dict) -> None:
        """
        Logs a prediction or feedback event.

        Args:
            application: Name of the application
            event: Data to logs as event body
        """
        self.log_batch(application, [event])

    def log_batch(self, application: str, events: List[dict]) -> None:
        pass

    def ping(self):
        pass

    def ready(self):
        pass


class APILogStore(BaseLogStore):
    BATCH_SIZE = 20
    BATCH_UPLOAD_TIMEOUT = 20  # seconds before retrying on upload

    def __init__(
        self,
        location: str,
        api_key: Optional[str] = None,
        send_in_background: bool = True,
        bypass_firehose: bool = False,
        consumer_factory=BatchConsumer,
    ):
        """
        Send logged events directly to the Gantry API.
        This is the recommended log store when using the local Gantry stack.

        Args:
            location (str): Gantry API host URI
            api_key (str, Optional): Gantry API Key, retrieved from the dashboard.
            send_in_background (bool, true by default): Whether to send events
                in a background thread.
            bypass_firehose (bool, false by default): Bypass firehose streaming
                and send directly to DB.
            consumer_factory: Used only for testing. Never use parameter in production.
        """
        self._api_client = APIClient(location, api_key=api_key)

        self._bypass_firehose = bypass_firehose

        self.num_consumer_threads: int = 1
        self.send_in_background = send_in_background
        self.queue: Queue = Queue()

        self.consumers = []
        self._consumer_factory = consumer_factory

        if self.send_in_background:
            # On program exit, allow the consumer thread to exit cleanly.
            # This prevents exceptions and a messy shutdown when the
            # interpreter is destroyed before the daemon thread finishes
            # execution.
            atexit.register(self._join)
            for _ in range(self.num_consumer_threads):
                consumer = self._consumer_factory(self.queue, self.consumer_func)
                self.consumers.append(consumer)
                consumer.start()

    def ping(self) -> bool:
        """
        Pings the API to check if it is up and running. Returns True if alive, else False.
        """
        try:
            # Cannot use /healthz/* endpoints as those will be answered by nginx
            # need to use /.
            # See https://linear.app/gantry/issue/ENG-2978/revisit-ping-in-sdk
            self._api_client.request("GET", "/api/ping", raise_for_status=True)
            return True
        except Exception as e:
            logger.error(f"Error during ping: {e}")
            return False

    def ready(self) -> bool:
        """
        Checks whether the API is ready to receive traffic with
        the provided API Key.
        """
        try:
            self._api_client.request("GET", "/api/v1/auth-ping", raise_for_status=True)
            return True
        except Exception as e:
            logger.error(f"Error during api key check: {e}")
            return False

    def consumer_func(self, batch):
        # Catch all errors and do not raise in order
        # for thread consumer to continue running.
        try:
            return self.send_batch_as_raw_events(batch, as_raw=True)
        except Exception as e:
            logger.error("Internal error sending batches: %s", e)

    @backoff.on_exception(
        backoff.constant,
        (requests.Timeout, requests.ConnectionError),
        interval=BACKOFF_INTERVAL_FOR_TIMEOUT_SECS,
        max_tries=BATCH_UPLOAD_MAX_TRIES,
        on_backoff=backoff_retry_handler,
        on_giveup=backoff_giveup_handler,
        jitter=None,
    )
    @backoff.on_exception(
        backoff.constant,
        GantryRequestException,
        max_tries=BATCH_UPLOAD_MAX_TRIES,
        interval=BACKOFF_INTERVAL_FOR_ERROR_CODE_SECS,
        on_backoff=backoff_retry_handler,
        on_giveup=backoff_giveup_handler,
        giveup=non_retry_error,
        jitter=None,
    )
    def send_batch_as_raw_events(self, batch: List, as_raw: bool = False) -> None:
        if as_raw:
            data = b",".join(batch)
            data = bytes('{"events": [', "utf8") + data + bytes("]}", "utf8")
        else:
            data = json.dumps({"events": batch}, cls=EventEncoder).encode("utf8")

        params = {}
        if self._bypass_firehose:
            params["bypass-firehose"] = "true"

        response = self._api_client.request(
            "POST",
            "/api/v1/ingest/raw",
            data=data,
            params=params,
            headers={"Content-Type": "application/json"},
            timeout=APILogStore.BATCH_UPLOAD_TIMEOUT,
            raise_for_status=True,
        )

        if response.get("response") != "ok":
            logger.error("Failed to log events. Response = %s", response)

    def log_batch(self, application: str, events: List[dict]) -> None:
        if not self.send_in_background:
            logger.info("Sending batch synchronously")
            batch_iterator_builder = _batch_iterator_factory(events, APILogStore.BATCH_SIZE)
            for batch in batch_iterator_builder():
                self.send_batch_as_raw_events(batch)
        else:
            for e in events:
                self.queue.put(e)

    def log_from_data_connector_async(
        self, request: IngestFromDataConnectorRequest
    ) -> IngestFromDataConnectorResponse:
        response = self._api_client.request(
            "POST",
            "/api/v1/ingest/data-connectors",
            json=asdict(request),
            headers={"Content-Type": "application/json"},
            raise_for_status=True,
        )

        return IngestFromDataConnectorResponse(**response["data"])

    def _join(self):
        """
        Ends the consumer thread once the queue is empty.
        Blocks execution until finished
        """
        for consumer in self.consumers:
            consumer.pause()
            try:
                consumer.join()
            except RuntimeError:
                # consumer thread has not started
                pass



import json
import random
import uuid

import mock
import pytest
import requests
import responses
from mock import patch
from responses import matchers

from gantry.exceptions import GantryRequestException
from gantry.logger.consumer import BatchConsumer
from gantry.logger.stores import (
    BATCH_UPLOAD_MAX_TRIES,
    APILogStore,
    _batch_iterator_factory,
    backoff_giveup_handler,
    backoff_retry_handler,
)
from gantry.serializers import EventEncoder

from ..conftest import EncodedTestingBatchIter, TestingBatchConsumer

TEST_HOST = "http://test-api"


def test_backoff_retry_handler_does_not_throw():
    backoff_retry_handler({})


def test_backoff_giveup_handler_does_not_throw():
    backoff_giveup_handler({"elapsed": 2.5, "tries": 2})


@pytest.mark.parametrize(
    ["collection", "batch_size", "expected"],
    [
        (list(range(0, 11)), 3, [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]),
        (list(range(0, 11)), 2, [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9], [10]]),
        (list(range(0, 11)), 9, [[0, 1, 2, 3, 4, 5, 6, 7, 8], [9, 10]]),
        (list(range(0, 11)), 1, [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9], [10]]),
        (list(range(0, 11)), 100, [list(range(0, 11))]),
        ([1], 100, [[1]]),
        ([1], 1, [[1]]),
    ],
)
def test_batch_factory(collection, batch_size, expected):
    assert list(_batch_iterator_factory(collection, batch_size)()) == expected


@pytest.mark.parametrize("batch_size", [0, -1, -5])
def test_batch_factory_error(batch_size):
    with pytest.raises(ValueError):
        _ = _batch_iterator_factory([1, 2, 3], batch_size)


@pytest.mark.parametrize("status_code", [400, 404, 500])
def test_ping_fail(status_code):
    store = APILogStore(TEST_HOST)
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, f"{TEST_HOST}/api/ping", status=status_code)
        assert store.ping() is False


@pytest.mark.parametrize("status_code", [200])
def test_ping_success(status_code):
    store = APILogStore(TEST_HOST)
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            f"{TEST_HOST}/api/ping",
            status=status_code,
            json={"response": "ok"},
        )
        assert store.ping() is True


@pytest.mark.parametrize("status_code", [400, 401, 404, 500])
def test_ready_fail(status_code):
    store = APILogStore(TEST_HOST)
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, f"{TEST_HOST}/api/v1/auth-ping", status=status_code)
        assert store.ready() is False


@pytest.mark.parametrize("status_code", [200])
def test_ready_success(status_code):
    store = APILogStore(TEST_HOST)
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            f"{TEST_HOST}/api/v1/auth-ping",
            status=status_code,
            json={"response": "ok"},
        )
        assert store.ready() is True


@pytest.mark.parametrize(
    ["data", "as_raw"], [(["foo", "bar", "baz"], False), ([b'"foo"', b'"bar"', b'"baz"'], True)]
)
def test_send_batch_as_raw_events(data, as_raw):
    store = APILogStore(TEST_HOST)

    content = {"events": ["foo", "bar", "baz"]}

    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            "{}/api/v1/ingest/raw".format(TEST_HOST),
            json={"response": "ok"},
            headers={"Content-Type": "application/json"},
            match=[matchers.json_params_matcher(content)],
        )

        store.send_batch_as_raw_events(data, as_raw=as_raw)


@pytest.mark.parametrize("error_t", [requests.Timeout, requests.ConnectionError])
@mock.patch("gantry.logger.stores.APIClient.request")
def test_backoff_for_send_batch_as_raw_events(mock_request, error_t, monkeypatch):
    """Test retry in timeouts and connection errors"""
    mock_request.side_effect = error_t
    # From https://github.com/litl/backoff/blob/a1e19d96a8bd6a4cfb44c88f42f0d8e0415998ac/tests/test_integration.py#L16 # noqa
    log = []

    def sleep(seconds):
        log.append(seconds)

    monkeypatch.setattr("time.sleep", sleep)

    store = APILogStore(TEST_HOST)
    try:
        store.send_batch_as_raw_events(["somedata"])
    except error_t:
        pass

    assert mock_request.call_count == BATCH_UPLOAD_MAX_TRIES
    assert log == [1, 1]


@mock.patch("gantry.logger.stores.APIClient.request")
def test_backoff_for_send_batch_as_raw_events_status_code(mock_request, monkeypatch):
    """Test retry on 502 requests"""
    mock_request.side_effect = GantryRequestException("url", 502, "msg")
    # From https://github.com/litl/backoff/blob/a1e19d96a8bd6a4cfb44c88f42f0d8e0415998ac/tests/test_integration.py#L16  # noqa
    log = []

    def sleep(seconds):
        log.append(seconds)

    monkeypatch.setattr("time.sleep", sleep)

    store = APILogStore(TEST_HOST)
    try:
        store.send_batch_as_raw_events(["somedata"])
    except GantryRequestException:
        pass

    assert mock_request.call_count == BATCH_UPLOAD_MAX_TRIES
    assert log == [5, 5]


@pytest.mark.parametrize("status_code", [400, 504, 404, 500])
@mock.patch("gantry.logger.stores.APIClient.request")
def test_backoff_for_send_batch_as_raw_events_fatal_error(mock_request, status_code):
    """On status_code != 502, we don't retry"""
    mock_request.side_effect = GantryRequestException("url", status_code, "msg")

    store = APILogStore(TEST_HOST)
    try:
        store.send_batch_as_raw_events(["somedata"])
    except GantryRequestException:
        pass

    assert mock_request.call_count == 1


@pytest.mark.parametrize("send_in_background", [False, True])
def test_api_log_store(send_in_background):
    with patch("gantry.logger.stores.APIClient") as mock_client:
        instance = mock_client.return_value
        instance.request.return_value = {"response": "ok"}

        consumer_factory = lambda q, f: BatchConsumer(  # noqa
            queue=q, func=f, batch_iter=EncodedTestingBatchIter
        )

        store = APILogStore(
            TEST_HOST, send_in_background=send_in_background, consumer_factory=consumer_factory
        )

        events = []
        for i in range(10):
            ev = {
                "event_id": uuid.uuid4(),
                "data": {
                    "a": random.randint(0, 100),
                    "b": "2",
                },
            }
            events.append(ev)
            store.log("simple_test", ev)

        store._join()

        num_events_sent = 0
        for i, call_args in enumerate(instance.request.call_args_list):
            assert call_args.args[0] == "POST"
            assert call_args.args[1] == "/api/v1/ingest/raw"
            assert call_args.kwargs["data"] == json.dumps(
                {"events": [events[i]]}, cls=EventEncoder
            ).encode("utf8")

            req_data = json.loads(call_args.kwargs["data"])
            num_events_sent += len(req_data["events"])

        assert num_events_sent == 10


@pytest.mark.parametrize("send_in_background", [False, True])
def test_api_log_store_batch(send_in_background):
    with patch("gantry.logger.stores.APIClient") as mock_client:
        instance = mock_client.return_value
        instance.request.return_value = {"response": "ok"}

        consumer_factory = lambda q, f: BatchConsumer(  # noqa
            queue=q, func=f, batch_iter=EncodedTestingBatchIter
        )

        store = APILogStore(
            TEST_HOST, send_in_background=send_in_background, consumer_factory=consumer_factory
        )

        events = []
        for i in range(10):
            ev = {
                "event_id": uuid.uuid4(),
                "data": {
                    "a": random.randint(0, 100),
                    "b": "2",
                },
            }
            events.append(ev)

        store.log_batch("application", events)
        store._join()

        store._join()

        num_events = 0

        if send_in_background:
            for i, call_args in enumerate(instance.request.call_args_list):
                assert call_args.args[0] == "POST"
                assert call_args.args[1] == "/api/v1/ingest/raw"
                assert call_args.kwargs["data"] == json.dumps(
                    {"events": [events[i]]}, cls=EventEncoder
                ).encode("utf8")
                num_events += 1

            assert num_events == 10

        else:
            assert len(instance.request.call_args_list) == 1
            call_args = instance.request.call_args_list[0]
            assert call_args.args[0] == "POST"
            assert call_args.args[1] == "/api/v1/ingest/raw"
            assert call_args.kwargs["data"] == json.dumps(
                {"events": events}, cls=EventEncoder
            ).encode("utf8")


@mock.patch("gantry.logger.stores.APILogStore.send_batch_as_raw_events")
def test_consumer_func(mock_send):
    mock_send.side_effect = RuntimeError()
    consumer_factory = lambda q, f: mock.Mock()  # noqa
    batch = mock.Mock()

    store = APILogStore(TEST_HOST, consumer_factory=consumer_factory)

    # This should not raise an exception
    store.consumer_func(batch)

    mock_send.assert_called_once_with(batch, as_raw=True)


def test_consumer_creation_for_background_store():
    store = APILogStore(TEST_HOST, send_in_background=True, consumer_factory=TestingBatchConsumer)

    assert len(store.consumers) == 1
    assert store.consumers[0].func == store.consumer_func



import collections
import datetime
import json
from collections import defaultdict
from distutils.util import strtobool
from typing import Any, Dict, List, Optional, Tuple, Union

import dateparser
import pandas as pd
from cachetools import LRUCache, TTLCache, cached

from gantry.api_client import APIClient
from gantry.query.core.constants import (
    MAX_SIZE_CACHE_B,
    MAX_TTL_CACHE_SECS,
    SCHEMA_TYPES,
    TIMEZONE,
)
from gantry.query.core.utils import _build_empty_df_with_schema

ORDER_ASCENDING = "ascending"
ORDER_DESCENDING = "descending"
DRUID_TIME_COLUMN_NAME = "__time"
TIME_COLUMN_NAME = "timestamp"


def build_base_query_params(start_time, end_time, version) -> Dict[str, Any]:
    return {
        "start_time": start_time,
        "end_time": end_time,
        "version": version,
    }


class QueryInfo(
    collections.namedtuple(
        "QueryParams", "application_node_id application version environment start_time end_time"
    )
):
    def get_base_query_params(self) -> Dict[str, Any]:
        return build_base_query_params(self.start_time, self.end_time, self.version)


@cached(TTLCache(maxsize=MAX_SIZE_CACHE_B, ttl=MAX_TTL_CACHE_SECS))
def _get_raw_data(api_client: APIClient, query_info: QueryInfo, dumped_json_data: str):
    """TTL-Caches data requests to API to make
    repeated fetching operations faster.
    """
    return api_client.request(
        "POST",
        "/api/v1/applications/{}/raw_data".format(query_info.application_node_id),
        params=query_info.get_base_query_params(),
        json=json.loads(
            dumped_json_data
        ),  # Need to receive dumped data because json dicts are not hashable
        raise_for_status=True,
    )


@cached(LRUCache(maxsize=MAX_SIZE_CACHE_B))
def _get_metadata(api_client, application_name, frozen_query_params):
    # Reconstruct the query parameters as a true dictionary here, if not cached
    query_params = {k: v for k, v in frozen_query_params}
    response = api_client.request(
        "GET",
        "/api/v1/models/" + application_name + "/schemas",
        params=query_params,
        raise_for_status=True,
    )
    return response["data"]


class GantryQueryFrame(object):
    def __init__(self, api_client: APIClient, query_info: QueryInfo) -> None:
        self.api_client: APIClient = api_client
        self.query_info: QueryInfo = query_info

    @property  # type: ignore
    def metadata(self):
        # Workaround to allow for hashing of query params for caching
        frozen_query_params = frozenset(self.query_info.get_base_query_params().items())
        return _get_metadata(self.api_client, self.query_info.application, frozen_query_params)

    @property
    def _dtypes(self):
        _dtypes = {}
        for schema_type in SCHEMA_TYPES:
            metadata = self.metadata.get(schema_type, [])
            for column in metadata:
                column_dtype = (
                    "category"
                    if column["node_metadata"]["dtype"].get(
                        "is_categorical"
                    )  # Prevent KeyError in some cases where column['node_metadata']['dtype'] == {}
                    and not column["dtype"] == "bool"
                    else column["dtype"]
                )
                _dtypes[column["name"]] = column_dtype

        return _dtypes

    def _get_dtype(self, column_name: str) -> str:
        return self._dtypes.get(column_name)

    @staticmethod
    def resolve_time_strings(
        start_time: Union[str, datetime.datetime],
        end_time: Union[str, datetime.datetime],
    ) -> Tuple[datetime.datetime, datetime.datetime]:
        # Convert time strings to unambiguous datetime objects.
        if isinstance(start_time, str):
            start_time_ret = dateparser.parse(start_time, settings=TIMEZONE)  # type: ignore
            if start_time_ret is None:
                raise ValueError("Invalid start_time: {}".format(start_time))
        else:
            start_time_ret = start_time
        if isinstance(end_time, str):
            end_time_ret = dateparser.parse(end_time, settings=TIMEZONE)  # type: ignore
            if end_time_ret is None:
                raise ValueError("Invalid end_time: {}".format(end_time))
        else:
            end_time_ret = end_time

        return start_time_ret, end_time_ret

    @staticmethod
    def _convert_to_bool_helper(x):
        if isinstance(x, bool):
            return x
        if x:
            return strtobool(str(x))
        return None

    def _prepare_pandas_dataframe(
        self,
        data: List[Dict[str, Any]],
        column: Optional[str] = None,
    ) -> pd.DataFrame:
        df = pd.DataFrame.from_records(data)
        if df.empty:
            return _build_empty_df_with_schema(self._dtypes)

        if column:
            if column not in df.columns:
                raise ValueError("Column {} not found in Gantry dataframe".format(column))

            df = (
                df[[DRUID_TIME_COLUMN_NAME, column]] if DRUID_TIME_COLUMN_NAME in df else df[column]
            )

        if isinstance(df, pd.Series):
            df = df.to_frame()

        if DRUID_TIME_COLUMN_NAME in df:
            df[DRUID_TIME_COLUMN_NAME] = pd.to_datetime(
                df[DRUID_TIME_COLUMN_NAME], unit="ms", utc=True
            )
            df = (
                df.rename(columns={DRUID_TIME_COLUMN_NAME: TIME_COLUMN_NAME})
                .set_index(TIME_COLUMN_NAME)
                .sort_index()
            )

        # Some Gantry types correspond directly to pd types; others will be parsed as objects.
        # Note that the timestamp, if applicable, is not a column, but rather the index itself.
        for col in df.columns:
            dtype = self._get_dtype(col)

            # Bool columns from druid can be "true," "false," or None.
            # To convert to pandas boolean type we must first
            # convert to python bool where we can.
            if dtype == "bool":
                df[col] = df[col].map(self._convert_to_bool_helper).astype(pd.BooleanDtype())

            if dtype == "str":
                df[col] = df[col].astype(pd.StringDtype())

            if dtype == "category":
                df[col] = df[col].astype(pd.CategoricalDtype())

            if dtype == "float":
                df[col] = df[col].astype(float)

            if dtype == "int":
                # if null are present, will become float
                df[col] = df[col].astype(int, errors="ignore")

        return df[df.columns.sort_values(ascending=True)]

    def _head(
        self,
        n: int = 5,
        column: Optional[str] = None,
        query_filters: Optional[List[dict]] = None,
        tags: Optional[dict] = None,
    ) -> pd.DataFrame:
        """
        Return the first n rows.
        This function returns the first n rows for the object based
        on position. It is useful for quickly testing if your object
        has the right type of data in it.

        Note: This does not populate the dataframe's GantrySeries as this would be a subset
        of the frame's expected timeseries data.
        """
        if not (n > 0):
            raise ValueError("n must be a positive int. Got {}".format(n))
        data = self._populate_raw_request(
            query_filters=query_filters,
            ascending=True,
            n=n,
            tags=tags,
        )
        return self._prepare_pandas_dataframe(data, column)

    def _tail(
        self,
        n: int = 5,
        column: Optional[str] = None,
        query_filters: Optional[List[dict]] = None,
        tags: Optional[dict] = None,
    ) -> pd.DataFrame:
        """
        Return the last n rows.
        This function returns the last n rows for the object based
        on position. It is useful for quickly testing if your object
        has the right type of data in it.

        Note: This does not populate the dataframe's GantrySeries as this would be a subset
        of the frame's expected timeseries data.
        """
        if not (n > 0):
            raise ValueError("n must be a positive int. Got {}".format(n))
        data = self._populate_raw_request(
            query_filters=query_filters,
            ascending=False,
            n=n,
            tags=tags,
        )
        return self._prepare_pandas_dataframe(data, column)

    def _fetch(
        self,
        column: Optional[str] = None,
        query_filters: Optional[List[dict]] = None,
        tags: Optional[dict] = None,
    ) -> pd.DataFrame:
        """
        Return all data in the query window.
        """
        data = self._populate_raw_request(
            query_filters=query_filters,
            ascending=True,
            n=None,
            tags=tags,
        )
        return self._prepare_pandas_dataframe(data, column)

    def _populate_raw_request(
        self,
        n: Optional[int] = None,
        ascending: bool = True,
        query_filters: Optional[List[dict]] = None,
        tags: Optional[dict] = None,
    ):
        json_data: Dict[str, Any] = defaultdict(lambda: {})

        if tags:
            json_data["tags"] = tags

        if query_filters and len(query_filters) > 0:
            json_data["filters"] = query_filters

        if self.query_info.environment:
            json_data["tags"]["env"] = self.query_info.environment

        if n:
            json_data["limit"] = n

        json_data["order"] = ORDER_ASCENDING if ascending else ORDER_DESCENDING
        json_data["include_join_id"] = True

        events = []
        has_next = True
        while has_next:
            response = _get_raw_data(self.api_client, self.query_info, json.dumps(json_data))
            events.extend(response["events"])
            to_fetch = n or response["metadata"]["count"]
            has_next = response["metadata"]["has_next"] and (len(events) < to_fetch)
            json_data["offset"] = len(events)

        # Invert list if ascending is False because
        # the ORDER_DESCENDING parameter will return the
        # events in reverse order.
        data = events if ascending else events[::-1]

        # It's possible to retrieve more than the requested number of events
        # when n is large and doesn't fit in the first request.
        return data[:n] if n else data



import json
import logging
import logging.handlers
from pathlib import Path
from queue import Empty
from typing import Dict, List

import pytest

from gantry.logger.client import Gantry
from gantry.logger.stores import APILogStore, BaseLogStore
from gantry.serializers import EventEncoder

TIMEOUT = 0.1


class TestingBatchIter:
    def __init__(self, queue, encode=False):
        self.queue = queue
        self.encode = encode

    def __iter__(self):
        return self

    def empty(self):
        return self.queue.empty()

    def __next__(self):
        try:
            if self.encode:
                return [
                    json.dumps(
                        self.queue.get(block=True, timeout=TIMEOUT), cls=EventEncoder
                    ).encode("utf8")
                ]

            return [self.queue.get(block=True, timeout=TIMEOUT)]
        except Empty:
            return []


class EncodedTestingBatchIter(TestingBatchIter):
    def __init__(self, queue):
        super().__init__(queue, True)


class TestingBatchIterWithEmptyFlag(TestingBatchIter):
    def __init__(self, queue, empty=True):
        super().__init__(queue, True)
        self._empty = empty

    def empty(self):
        return self._empty


class TestingBatchConsumer:
    def __init__(self, queue, func):
        self.queue = queue
        self.func = func

        self.running = None

    def pause(self):
        self.running = False

    def start(self):
        self.running = True

    def join(self):
        pass


class FileLogStore(BaseLogStore):
    """This implementation is just for testing purposes"""

    _LOGGER_NS = "gantry.stores.file_logger"
    MAX_FILE_SIZE = 1000000 * 10  # 10MB

    def __init__(self, location: str):
        self._location = Path(location).resolve()
        self._loggers: Dict[str, logging.Logger] = {}

        self._location.mkdir(parents=True, exist_ok=True)

    def _init_logger(self, prefix: str, name: str) -> logging.Logger:
        # TODO: make sure name is a valid file name
        logdir = self._location / prefix
        logdir.mkdir(parents=True, exist_ok=True)

        filepath = logdir / "{}.log".format(name)
        event_logger = logging.getLogger("{}.{}.{}".format(self._LOGGER_NS, prefix, name))
        event_logger.setLevel(logging.INFO)
        # Makes sure event_logger events are not output by the gantry logger
        event_logger.propagate = False
        match_existing_handler = [
            h
            for h in event_logger.handlers
            if (isinstance(h, logging.FileHandler) and h.baseFilename == str(filepath))
        ]
        if not match_existing_handler:
            ch = logging.handlers.RotatingFileHandler(str(filepath), maxBytes=self.MAX_FILE_SIZE)
            formatter = logging.Formatter("%(message)s")
            ch.setFormatter(formatter)
            event_logger.addHandler(ch)
        return event_logger

    def log_batch(self, name: str, events: List[dict], batch_meta: dict = None) -> None:
        for e in events:
            key: str = self._get_key(name, e)
            self._loggers[key].info(json.dumps(e, cls=EventEncoder))

    def _get_key(self, name: str, event: dict) -> str:
        has_feedback = "feedback" in event
        has_predictions = "inputs" in event
        prefix = ""

        if has_feedback and has_predictions:
            prefix = "records"
        elif has_feedback:
            prefix = "feedback"
        elif has_predictions:
            prefix = "predictions"

        key = "{}/{}".format(prefix, name)

        if key not in self._loggers:
            self._loggers[key] = self._init_logger(prefix, name)

        return key


@pytest.fixture(scope="function")
def gantry(tmp_path):
    import gantry.logger.main as gantry_module

    gantry_module._CLIENT = None

    yield gantry_module

    gantry_module._CLIENT = None


@pytest.fixture(scope="function")
def initialized_gantry(tmp_path):
    import gantry.logger.main as gantry_module

    gantry_module._CLIENT = None

    log_store = FileLogStore(str(tmp_path))

    gantry_module._CLIENT = Gantry(log_store, environment="test")

    yield gantry_module

    gantry_module._CLIENT = None


@pytest.fixture(scope="function")
def init_gantry_with_api():
    import gantry.logger.main as gantry_module

    gantry_module._CLIENT = None

    log_store = APILogStore(location="http://test-api", send_in_background=False)
    gantry_module._CLIENT = Gantry(log_store, environment="test")

    yield gantry_module

    gantry_module._CLIENT = None


@pytest.fixture(scope="function")
def config_data(tmp_path):
    return {"logs_location": str(tmp_path)}



import pytest
import datetime
import pandas.api.types as ptypes
from gantry.data_generator.data_generator import generate_demo_data, _generate_batch_data


@pytest.mark.parametrize(
    "sample_size,expected_length",
    [
        (203, 203),
        (500, 500),
        (1000, 1000),
        (-2000, 200),
        (-200, 200),
        (-1, 200),
        (0, 200),
        (50, 200),
        (5000, 4000),
    ],
)
def test_generate_demo_data(sample_size, expected_length):
    data = generate_demo_data(sample_size)
    assert len(data) == expected_length
    assert ptypes.is_numeric_dtype(data["income"])
    assert ptypes.is_string_dtype(data["description"])
    assert all(ptypes.is_bool_dtype(data[col]) for col in ["prediction", "loan_repaid"])
    assert ptypes.is_categorical_dtype(data["credit_history"])
    assert ptypes.is_datetime64_any_dtype(data["timestamp"])
    assert max(data["timestamp"]) <= datetime.datetime.now()
    assert min(data["income"]) >= 5000


@pytest.mark.parametrize(
    "cred_hist_prop,pred_true_rate,loan_repaid_rate",
    [
        (-0.5, 0.2, 0.2),
        (0.2, -0.5, 0.2),
        (0.2, 0.2, -0.5),
        (0.2, 0.2, 2),
        (0.2, 2, 0.2),
        (2, 0.2, 0.2),
    ],
)
def test_generate_batch_data(cred_hist_prop, pred_true_rate, loan_repaid_rate):
    with pytest.raises(ValueError):
        _generate_batch_data(
            cred_hist_prop=cred_hist_prop,
            pred_true_rate=pred_true_rate,
            loan_repaid_rate=loan_repaid_rate,
        )



from typing import Optional

import click
from click_aliases import ClickAliasedGroup

import gantry
from gantry.query.core.utils import get_application_node_id


@click.group(cls=ClickAliasedGroup)
def application():
    """
    Use this to handle application metadata in gantry.
    """


@application.command(
    aliases=["hard_delete"],
    help="Hard delete an existing application in Gantry. \
    This is an irreversible action!",
)
@click.option("--application-name", type=click.STRING)
@click.option("--version", default=None, type=click.STRING)
def hard_delete(application_name: str, version: Optional[str]):
    gantry.init()
    api_client = gantry.get_client().log_store._api_client
    click.confirm(
        f"Are you sure you want to hard delete the application: {application_name} \
at version: {version if version is not None else 'latest'}?",
        abort=True,
    )
    application_id = get_application_node_id(api_client, application_name, version=version)
    response = api_client.request(
        "POST",
        f"/api/v1/applications/{application_id}/hard_delete",
    )
    if response.get("response") != "ok":
        err_msg = response.get("error", "unknown error")
        raise click.ClickException(f"Fail to hard delete application: {err_msg}.")



import os

import mock
import pytest

import gantry
from gantry.exceptions import ClientNotInitialized
from gantry.logger import main


@pytest.mark.parametrize(
    ["method", "params"],
    [
        ("ping", {}),
        ("ready", {}),
        ("log_record", {"application": "foobar", "version": "1.2.3"}),
        ("log_records", {"application": "foobar", "version": "1.2.3"}),
    ],
)
@pytest.mark.parametrize("module", [main, gantry])
def test_uninit_client_main(module, method, params):
    """Test all public methods from gantry module or gantry.main module
    fail if client is not initialized
    """
    with mock.patch("gantry.logger.main._CLIENT", None):
        with pytest.raises(ClientNotInitialized):
            getattr(module, method)(**params)


@pytest.mark.parametrize(
    ["method", "params"],
    [
        ("ping", {}),
        ("ready", {}),
        (
            "log_records",
            {
                "application": "foobar",
                "version": "1.2.3",
                "inputs": [1, 2, 3],
                "outputs": [4, 5, 6],
                "feedback_keys": ["A"],
                "feedback_ids": [10],
                "feedbacks": [4, 5, 6],
                "ignore_inputs": ["A"],
                "timestamps": "today",
                "sort_on_timestamp": True,
                "sample_rate": 1.0,
                "as_batch": False,
                "tags": {},
                "join_keys": ["foo", "bar"],
                "row_tags": [{}, {}, {}],
                "global_tags": {},
            },
        ),
        (
            "log_record",
            {
                "application": "foobar",
                "version": "1.2.3",
                "inputs": [1, 2, 3],
                "outputs": [4, 5, 6],
                "feedback_keys": ["A"],
                "feedback_id": [10],
                "feedback": [4, 5, 6],
                "ignore_inputs": ["A"],
                "timestamp": "today",
                "sample_rate": 1.0,
                "tags": {},
                "join_key": "foo",
            },
        ),
    ],
)
@pytest.mark.parametrize("module", [main, gantry])
def test_logger_methods(module, method, params):
    """Test all public methods from gantry module or gantry.main module
    resolve in the global _CLIENT methods
    """
    m = mock.Mock()
    with mock.patch("gantry.logger.main._CLIENT", m):
        getattr(m, method).return_value = "return_value"
        assert getattr(module, method)(**params) == "return_value"
        getattr(m, method).assert_called_once_with(**params)


@pytest.mark.parametrize(
    ["kwargs", "expected"],
    [
        (
            {"api_key": "12345"},
            ("12345", None, None, None),
        ),
        (
            {
                "api_key": "12345",
                "logging_level": "ERROR",
                "environment": "test",
                "send_in_background": False,
            },
            ("12345", "ERROR", "test", False),
        ),
    ],
)
@pytest.mark.parametrize("ping_ret", [True, False])
@mock.patch("gantry.logger.main.Gantry")
@mock.patch("gantry.logger.main._resolve_params")
def test_logger_init(mock_resolve_params, mock_gantry, ping_ret, kwargs, expected):
    """Test the initialzation flow"""
    mock_resolve_params.return_value = (
        "54321",
        "https://barbaz",
        "CRITICAL",
        "foobaz",
        False,
        False,
    )
    m = mock.Mock()
    m.ping = mock.Mock(return_value=ping_ret)
    mock_gantry.return_value = m

    main._init(**kwargs)

    assert main._CLIENT == m

    args, kwargs = mock_gantry.call_args
    store = args[0]
    assert store._api_client._host == "https://barbaz"
    assert store._api_client._api_key == "54321"
    assert store._bypass_firehose is False
    assert store.send_in_background is False

    assert args[1] == "foobaz"
    assert args[2] == "CRITICAL"

    mock_resolve_params.assert_called_once_with(*expected)
    m.ping.assert_called_once()
    if ping_ret:
        m.ready.assert_called_once()


@pytest.mark.parametrize(
    ["environ", "kwargs", "expected"],
    [
        (
            {},
            {
                "api_key": "12345",
                "logging_level": "ERROR",
                "environment": "test",
                "send_in_background": False,
            },
            ("12345", "https://app.gantry.io", "ERROR", "test", False, False),
        ),
        (
            {
                "GANTRY_API_KEY": "123456789",
                "GANTRY_LOGS_LOCATION": "https://barfoo",
                "GANTRY_LOGGING_LEVEL": "WARNING",
                "GANTRY_ENVIRONMENT": "another",
                "GANTRY_BYPASS_FIREHOSE": "true",
                "GANTRY_SEND_IN_BACKGROUND": "true",
            },
            {
                "api_key": "12345",
                "logging_level": "ERROR",
                "environment": "test",
                "send_in_background": False,
            },
            ("12345", "https://barfoo", "ERROR", "test", True, False),
        ),
        (
            {
                "GANTRY_API_KEY": "123456789",
                "GANTRY_LOGGING_LEVEL": "WARNING",
                "GANTRY_ENVIRONMENT": "another",
                "GANTRY_SEND_IN_BACKGROUND": "true",
            },
            {
                "logging_level": "ERROR",
                "environment": "test",
            },
            ("123456789", "https://app.gantry.io", "ERROR", "test", False, True),
        ),
        (
            {
                "GANTRY_API_KEY": "123456789",
            },
            {},
            ("123456789", "https://app.gantry.io", "INFO", "dev", False, True),
        ),
        (
            {
                "GANTRY_API_KEY": "123456789",
                "AWS_LAMBDA_FUNCTION_NAME": "foobar",
            },
            {},
            ("123456789", "https://app.gantry.io", "INFO", "dev", False, False),
        ),
    ],
)
def test_resolve_params(environ, kwargs, expected):
    """Test that passed parameters take precedence over env vars"""
    with mock.patch.dict(os.environ, environ):
        assert expected == main._resolve_params(**kwargs)


@pytest.mark.parametrize(
    ["environ", "kwargs"],
    [
        (
            {},
            {
                "logs_location": "https://foobar",
                "logging_level": "ERROR",
                "environment": "test",
                "bypass_firehose": False,
                "send_in_background": False,
            },
        ),
        (
            {
                "GANTRY_API_KEY": "123456789",
                "GANTRY_LOGS_LOCATION": "https://barfoo",
                "GANTRY_LOGGING_LEVEL": "WARNING",
                "GANTRY_ENVIRONMENT": "another",
                "GANTRY_BYPASS_FIREHOSE": "true",
                "GANTRY_SEND_IN_BACKGROUND": "true",
            },
            {
                "api_key": "12345",
                "logs_location": "tcp://foobar",
                "logging_level": "ERROR",
                "environment": "test",
                "bypass_firehose": False,
                "send_in_background": False,
            },
        ),
        (
            {
                "GANTRY_API_KEY": "123456789",
                "GANTRY_LOGS_LOCATION": "https://barfoo",
            },
            {
                "logs_location": "https://foobar",
                "logging_level": "another",
            },
        ),
    ],
)
def test_resolve_params_invalid(environ, kwargs):
    """Check invalid initialization parameters even env vars"""
    with mock.patch.dict(os.environ, environ):
        with pytest.raises((ValueError, TypeError)):
            main._resolve_params(**kwargs)



import mock
import pytest
from click.testing import CliRunner

from gantry.cli.projection import _get_custom_projection_configs, get_logs, update


def test_get_custom_projection_configs():
    empty_projection_dir = "tests/cli/custom-projections/empty_projection_dir/"
    with pytest.raises(ValueError):
        _get_custom_projection_configs(empty_projection_dir)

    valid_projection_dir = "tests/cli/custom-projections/valid_projection_dir/"
    projection = _get_custom_projection_configs(valid_projection_dir)
    assert projection is not None


@mock.patch(
    "gantry.cli.projection._validate_custom_projection_configs",
    return_value=False,
)
@mock.patch("gantry.logger.main.get_client")
def test_update_fail_validate_configs(mock_api_client, mock_validate_configs):
    runner = CliRunner()
    cli_args = ["--projection-dir", "tests/cli/custom-projections/invalid_projection_dir/"]
    result = runner.invoke(update, cli_args, env={"GANTRY_API_KEY": "test"})
    assert result.exit_code == 1
    assert (
        "Validate custom projection definition...FAILED.\nError: Invalid projection definition\n"
        in result.output
    )


@mock.patch(
    "gantry.cli.projection._validate_custom_projection_configs",
    return_value=(True, []),
)
@mock.patch("gantry.logger.main.get_client")
@mock.patch(
    "gantry.cli.projection._get_s3_destination",
    return_value=(False, {"response": "error", "error": "S3 error"}),
)
def test_update_fail_get_s3_destination(mock_s3, mock_api_client, mock_validate_configs):
    runner = CliRunner()
    cli_args = ["--projection-dir", "tests/cli/custom-projections/valid_projection_dir/"]
    result = runner.invoke(update, cli_args, env={"GANTRY_API_KEY": "test"})
    assert result.exit_code == 1
    assert "Fail to set up upload location: S3 error" in result.output


@mock.patch(
    "gantry.cli.projection._validate_custom_projection_configs",
    return_value=(True, []),
)
@mock.patch("gantry.cli.projection.APIClient.request")
@mock.patch(
    "gantry.cli.projection._get_s3_destination",
    return_value=(True, {"response": "ok", "upload_url": "http://test_url", "s3_key": "test_key"}),
)
@mock.patch("gantry.cli.projection.subprocess")
@mock.patch(
    "gantry.cli.projection._upload_function_zip",
    return_value=(True, ""),
)
@mock.patch(
    "gantry.cli.projection._get_build_logs",
    return_value=("ERROR", "Error getting task logs"),
)
def test_update_success_error_get_logs(
    mock_logs, mock_zip, mock_subprocess, mock_s3, mock_api_client, mock_validate_configs
):
    mock_subprocess.run.return_value.returncode = 0
    mock_api_client.return_value = {"response": "ok", "task_id": "123abc"}

    runner = CliRunner()
    cli_args = ["--projection-dir", "tests/cli/custom-projections/valid_projection_dir/"]
    result = runner.invoke(update, cli_args, env={"GANTRY_API_KEY": "test"})
    assert result.exit_code == 0
    assert "SUCCESS\n--> Task 123abc has been submitted." in result.output


@mock.patch("gantry.cli.projection.APIClient.request")
def test_get_logs_failure(mock_api_client):
    mock_api_client.return_value = {"response": "error", "error": "some error", "logs": []}

    runner = CliRunner()
    cli_args = [
        "--application-name",
        "test-app",
        "--projection-name",
        "test-projection",
        "--start",
        "2022-11-15T17:00:00",
        "--end",
        "2022-11-15T17:01:00",
    ]
    result = runner.invoke(get_logs, cli_args, env={"GANTRY_API_KEY": "test"})
    assert result.exit_code == 1
    assert "Fail to retrieve logs: some error" in result.output


@mock.patch("gantry.cli.projection.APIClient.request")
def test_get_logs_success(mock_api_client):
    mock_api_client.return_value = {
        "response": "ok",
        "logs": [
            {
                "eventId": "123",
                "ingestionTime": 166876447801,
                "logStreamName": "stream",
                "message": "START RequestId: 2c0c8dd2-de6b-4e5c-a7aa-b8ecf53cafb2 Version: 1\n",
                "timestamp": 166876447800,
            },
            {
                "eventId": "456",
                "ingestionTime": 166876447802,
                "logStreamName": "stream",
                "message": "REPORT RequestId: 2c0c8dd2-de6b-4e5c-a7aa-b8ecf53cafb2\tDuration: 0.99 ms\tBilled Duration: 1 ms\tMemory Size: 128 MB\tMax Memory Used: 37 MB\t\n",  # noqa
                "timestamp": 166876447801,
            },
            {
                "eventId": "789",
                "ingestionTime": 166876447803,
                "logStreamName": "stream",
                "message": "END RequestId: 2c0c8dd2-de6b-4e5c-a7aa-b8ecf53cafb2\n",
                "timestamp": 166876447802,
            },
        ],
    }

    runner = CliRunner()
    cli_args = [
        "--application-name",
        "test-app",
        "--projection-name",
        "test-projection",
        "--start",
        "2022-11-15T17:00:00",
        "--end",
        "2022-11-15T17:01:00",
    ]
    result = runner.invoke(get_logs, cli_args, env={"GANTRY_API_KEY": "test"})
    assert result.exit_code == 0



import logging
import platform
from typing import Optional
from urllib.parse import urljoin

import requests

from gantry.exceptions import GantryRequestException
from gantry.serializers import dict_encoder
from gantry.version import __version__

logger = logging.getLogger(__name__)

API_KEY_HEADER_NAME = "X-Gantry-Api-Key"


class APIClient:
    def __init__(self, origin: str, api_key: Optional[str] = None):
        self._host = origin
        self._api_key = api_key

        user_agent = "gantry/{} {}/{} {} {}".format(
            __version__,
            platform.python_implementation(),
            platform.python_version(),
            platform.system(),
            platform.machine(),
        )

        self._session = requests.Session()
        self._session.headers.update({"User-Agent": user_agent})

    def request(
        self,
        method,
        path,
        params=None,
        data=None,
        files=None,
        headers=None,
        timeout=None,
        proxies=None,
        verify=None,
        json=None,
        raise_for_status: bool = False,
        raw_response: bool = False,
    ):
        """
        If raise_for_status is True, the method will raise GantryRequestException if the response
        status code is >= 400
        """
        url = urljoin(self._host, path)

        req_headers = {}
        if self._api_key:
            req_headers[API_KEY_HEADER_NAME] = self._api_key
        if headers:
            req_headers.update(headers)

        if json:
            json = dict_encoder(json)

        logger.debug("%s %s", method, url)

        response = self._session.request(
            method,
            url,
            params=params,
            data=data,
            files=files,
            headers=req_headers,
            timeout=timeout,
            proxies=proxies,
            verify=verify,
            json=json,
        )

        logger.debug("Response: [%s]", response.status_code)
        logger.debug(response.text.strip("\n"))  # Remove newlines not to mess up with tracing tools

        if raise_for_status and response.status_code >= 400:
            raise GantryRequestException(url, response.status_code, response.text)

        try:
            return response.json() if not raw_response else response
        except Exception as e:
            logger.debug(f"Invalid json response: {response.text} [{str(e)}]")
            return {"response": response.status_code, "error": response.status_code}


class LocalClient(APIClient):
    """For testing purposes only"""

    def __init__(self, app, api_key=None):
        self._app = app
        self._api_key = api_key

    def request(
        self,
        method,
        path,
        params=None,
        data=None,
        headers=None,
        timeout=None,
        proxies=None,
        verify=None,
        json=None,
    ):
        if json:
            json = dict_encoder(json)

        req_headers = {}
        if self._api_key:
            req_headers[API_KEY_HEADER_NAME] = self._api_key
        if headers:
            req_headers.update(headers)

        with self._app.test_client() as c:
            response = c.open(
                method=method,
                path=path,
                query_string=params,
                data=data,
                headers=req_headers,
                json=json,
            )

            # TODO: add common error handling here
            return response.get_json()



import mock

from gantry import init


@mock.patch("gantry.logger_init")
@mock.patch("gantry.query_init")
@mock.patch("gantry.alerts_init")
@mock.patch("gantry.curators_init")
@mock.patch("gantry.dataset_init")
def test_init(
    mock_dataset_init, mock_curators_init, mock_alerts_init, mock_query_init, mock_logger_init
):
    init("foobar", "INFO", "dev", False)

    mock_dataset_init.assert_called_once_with("foobar")
    mock_curators_init.assert_called_once_with("foobar")
    mock_alerts_init.assert_called_once_with("foobar")
    mock_query_init.assert_called_once_with("foobar")
    mock_logger_init.assert_called_once_with("foobar", "INFO", "dev", False)
